from __future__ import annotations

import json
import logging
import os
import re
import sys
import tempfile
import uuid

from hashlib import sha256
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

import pytest

from packaging.utils import canonicalize_name
from poetry.core.constraints.version import Version
from poetry.core.packages.dependency_group import MAIN_GROUP
from poetry.core.packages.package import Package
from poetry.core.packages.project_package import ProjectPackage
from poetry.core.version.markers import AnyMarker
from poetry.core.version.markers import parse_marker

from poetry.__version__ import __version__
from poetry.factory import Factory
from poetry.packages.locker import GENERATED_COMMENT
from poetry.packages.locker import Locker
from poetry.packages.transitive_package_info import TransitivePackageInfo
from tests.helpers import get_dependency
from tests.helpers import get_package


if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture


DEV_GROUP = canonicalize_name("dev")


@pytest.fixture
def locker() -> Locker:
    with tempfile.NamedTemporaryFile() as f:
        f.close()
        locker = Locker(Path(f.name), {})

        return locker


@pytest.fixture
def root() -> ProjectPackage:
    return ProjectPackage("root", "1.2.3")


@pytest.fixture
def transitive_info() -> TransitivePackageInfo:
    return TransitivePackageInfo(0, {MAIN_GROUP}, {})


@pytest.mark.parametrize("is_locked", [True, False])
def test_is_locked(locker: Locker, root: ProjectPackage, is_locked: bool) -> None:
    if is_locked:
        locker.set_lock_data(root, {})
    assert locker.is_locked() is is_locked


@pytest.mark.parametrize("is_fresh", [True, False])
def test_is_fresh(
    locker: Locker,
    root: ProjectPackage,
    transitive_info: TransitivePackageInfo,
    is_fresh: bool,
) -> None:
    locker.set_lock_data(root, {})
    if not is_fresh:
        locker.set_pyproject_data(
            {"tool": {"poetry": {"dependencies": {"tomli": "*"}}}}
        )
    assert locker.is_fresh() is is_fresh


@pytest.mark.parametrize(
    ("kind", "version", "expected"),
    [
        ("valid", "2.3.0", True),
        ("legacy", "2.3.0", False),
        ("outdated", "2.3.0", False),
        ("valid", "2.2.1", True),
        ("legacy", "2.2.1", True),
        ("outdated", "2.3.0", False),
    ],
)
def test_is_fresh_dependency_groups(
    locker: Locker,
    root: ProjectPackage,
    transitive_info: TransitivePackageInfo,
    kind: Literal["valid", "legacy", "outdated"],
    version: str,
    expected: bool,
) -> None:
    locker.set_lock_data(root, {})
    locker.set_pyproject_data({"dependency-groups": {"foo": []}})
    if kind == "valid":
        locked_hash = locker._get_content_hash()
    elif kind == "legacy":
        locked_hash = locker._get_content_hash(with_dependency_groups=False)
        assert locked_hash != locker._get_content_hash()
    else:
        locked_hash = "123456"

    lock_content = locker.lock.read_text(encoding="utf-8")
    lock_content = re.sub(r"Poetry [^ ]+", f"Poetry {version}", lock_content)
    lock_content = re.sub(
        r'content-hash = "[^"]+"', f'content-hash = "{locked_hash}"', lock_content
    )
    locker.lock.write_text(lock_content, encoding="utf-8")

    assert locker.is_fresh() is expected


@pytest.mark.parametrize("lock_version", [None, "2.0", "2.1"])
def test_is_locked_group_and_markers(
    locker: Locker, root: ProjectPackage, lock_version: str | None
) -> None:
    if lock_version:
        locker.set_lock_data(root, {})
        with locker.lock.open("r", encoding="utf-8") as f:
            content = f.read()
        content = content.replace(locker._VERSION, lock_version)
        with locker.lock.open("w", encoding="utf-8") as f:
            f.write(content)
    assert locker.is_locked_groups_and_markers() is (lock_version == "2.1")


def test_lock_file_data_is_ordered(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_a = get_package("A", "1.0.0")
    package_a.add_dependency(Factory.create_dependency("B", "^1.0"))
    package_a.files = [{"file": "foo", "hash": "456"}, {"file": "bar", "hash": "123"}]
    package_a2 = get_package("A", "2.0.0")
    package_a2.files = [{"file": "baz", "hash": "345"}]
    package_git = Package(
        "git-package",
        "1.2.3",
        source_type="git",
        source_url="https://github.com/python-poetry/poetry.git",
        source_reference="develop",
        source_resolved_reference="123456",
    )
    package_git_with_subdirectory = Package(
        "git-package-subdir",
        "1.2.3",
        source_type="git",
        source_url="https://github.com/python-poetry/poetry.git",
        source_reference="develop",
        source_resolved_reference="123456",
        source_subdirectory="subdir",
    )
    package_url_linux = Package(
        "url-package",
        "1.0",
        source_type="url",
        source_url="https://example.org/url-package-1.0-cp39-manylinux_2_17_x86_64.whl",
    )
    package_url_win32 = Package(
        "url-package",
        "1.0",
        source_type="url",
        source_url="https://example.org/url-package-1.0-cp39-win_amd64.whl",
    )
    packages = {
        package_a2: transitive_info,
        package_a: transitive_info,
        get_package("B", "1.2"): transitive_info,
        package_git: transitive_info,
        package_git_with_subdirectory: transitive_info,
        package_url_win32: transitive_info,
        package_url_linux: transitive_info,
    }

    locker.set_lock_data(root, packages)

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = [
    {{file = "bar", hash = "123"}},
    {{file = "foo", hash = "456"}},
]

[package.dependencies]
B = "^1.0"

[[package]]
name = "A"
version = "2.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = [
    {{file = "baz", hash = "345"}},
]

[[package]]
name = "B"
version = "1.2"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[[package]]
name = "git-package"
version = "1.2.3"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []
develop = false

[package.source]
type = "git"
url = "https://github.com/python-poetry/poetry.git"
reference = "develop"
resolved_reference = "123456"

[[package]]
name = "git-package-subdir"
version = "1.2.3"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []
develop = false

[package.source]
type = "git"
url = "https://github.com/python-poetry/poetry.git"
reference = "develop"
resolved_reference = "123456"
subdirectory = "subdir"

[[package]]
name = "url-package"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.source]
type = "url"
url = "https://example.org/url-package-1.0-cp39-manylinux_2_17_x86_64.whl"

[[package]]
name = "url-package"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.source]
type = "url"
url = "https://example.org/url-package-1.0-cp39-win_amd64.whl"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_properly_loads_extras(locker: Locker) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "cachecontrol"
version = "0.12.5"
description = "httplib2 caching for requests"
optional = false
python-versions = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*"
groups = ["main"]
files = []

[package.dependencies]
msgpack = "*"
requests = "*"

[package.dependencies.lockfile]
optional = true
version = ">=0.9"

[package.extras]
filecache = ["lockfile (>=0.9)"]
redis = ["redis (>=2.10.5)"]

[metadata]
lock-version = "2.1"
python-versions = "~2.7 || ^3.4"
content-hash = "c3d07fca33fba542ef2b2a4d75bf5b48d892d21a830e2ad9c952ba5123a52f77"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    packages = locker.locked_repository().packages

    assert len(packages) == 1

    package = packages[0]
    assert len(package.requires) == 3
    assert len(package.extras) == 2

    lockfile_dep = package.extras[canonicalize_name("filecache")][0]
    assert lockfile_dep.name == "lockfile"


def test_locker_properly_loads_nested_extras(locker: Locker) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "a"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
b = {{version = "^1.0", optional = true, extras = "c"}}

[package.extras]
b = ["b[c] (>=1.0,<2.0)"]

[[package]]
name = "b"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
c = {{version = "^1.0", optional = true}}

[package.extras]
c = ["c (>=1.0,<2.0)"]

[[package]]
name = "c"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[metadata]
python-versions = "*"
lock-version = "2.1"
content-hash = "123456789"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    repository = locker.locked_repository()
    assert len(repository.packages) == 3

    packages = repository.find_packages(get_dependency("a", "1.0"))
    assert len(packages) == 1

    package = packages[0]
    assert len(package.requires) == 1
    assert len(package.extras) == 1

    dependency_b = package.extras[canonicalize_name("b")][0]
    assert dependency_b.name == "b"
    assert dependency_b.extras == frozenset({"c"})

    packages = repository.find_packages(dependency_b)
    assert len(packages) == 1

    package = packages[0]
    assert len(package.requires) == 1
    assert len(package.extras) == 1

    dependency_c = package.extras[canonicalize_name("c")][0]
    assert dependency_c.name == "c"
    assert dependency_c.extras == frozenset()

    packages = repository.find_packages(dependency_c)
    assert len(packages) == 1


def test_locker_properly_loads_extras_legacy(locker: Locker) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "a"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
b = {{version = "^1.0", optional = true}}

[package.extras]
b = ["b (^1.0)"]

[[package]]
name = "b"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[metadata]
python-versions = "*"
lock-version = "2.1"
content-hash = "123456789"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    repository = locker.locked_repository()
    assert len(repository.packages) == 2

    packages = repository.find_packages(get_dependency("a", "1.0"))
    assert len(packages) == 1

    package = packages[0]
    assert len(package.requires) == 1
    assert len(package.extras) == 1

    dependency_b = package.extras[canonicalize_name("b")][0]
    assert dependency_b.name == "b"


def test_locker_properly_loads_subdir(locker: Locker) -> None:
    content = """\
[[package]]
name = "git-package-subdir"
version = "1.2.3"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false
files = []

[package.source]
type = "git"
url = "https://github.com/python-poetry/poetry.git"
reference = "develop"
resolved_reference = "123456"
subdirectory = "subdir"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""
    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    repository = locker.locked_repository()
    assert len(repository.packages) == 1

    packages = repository.find_packages(get_dependency("git-package-subdir", "1.2.3"))
    assert len(packages) == 1

    package = packages[0]
    assert package.source_subdirectory == "subdir"


@pytest.mark.parametrize(
    ("groups", "marker", "expected"),
    [
        # only main - without marker
        (["main"], None, {"main": "*"}),
        # only main - with marker
        (
            ["main"],
            repr('python_version == "3.9"'),
            {"main": 'python_version == "3.9"'},
        ),
        # two groups - common marker
        (
            ["main", "dev"],
            repr('python_version == "3.9"'),
            {"main": 'python_version == "3.9"', "dev": 'python_version == "3.9"'},
        ),
        # two groups - separate marker
        (
            ["main", "dev"],
            (
                '{"main" = \'python_version == "3.9"\','
                ' "dev" = \'sys_platform == "linux"\'}'
            ),
            {"main": 'python_version == "3.9"', "dev": 'sys_platform == "linux"'},
        ),
        # two groups - one without marker
        (
            ["main", "dev"],
            '{"main" = \'python_version == "3.9"\'}',
            {"main": 'python_version == "3.9"', "dev": "*"},
        ),
        (
            # unnormalized group - common marker
            ["main", "DEV"],
            repr('python_version == "3.9"'),
            {"main": 'python_version == "3.9"', "dev": 'python_version == "3.9"'},
        ),
        (
            # unnormalized group - separate marker
            ["main", "DEV"],
            (
                '{"main" = \'python_version == "3.9"\','
                ' "DEV" = \'sys_platform == "linux"\'}'
            ),
            {"main": 'python_version == "3.9"', "dev": 'sys_platform == "linux"'},
        ),
    ],
)
def test_locker_properly_loads_groups_and_markers(
    locker: Locker, groups: list[str], marker: str, expected: dict[str, str]
) -> None:
    content = rf"""
[[package]]
name = "a"
version = "1.0"
optional = false
python-versions = "*"
groups = {groups}
{"markers = " + marker if marker else ""}
files = []

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""
    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    packages = locker.locked_packages()

    a = get_package("a", "1.0")
    assert len(packages) == 1
    assert packages[a].groups == {canonicalize_name(g) for g in groups}
    assert packages[a].markers == {g: parse_marker(m) for g, m in expected.items()}


def test_locker_properly_assigns_metadata_files(locker: Locker) -> None:
    """
    For multiple constraints dependencies, there is only one common entry in
    metadata.files. However, we must not assign all the files to each of the packages
    because this can result in duplicated and outdated entries when running
    `poetry lock` and hash check failures when running `poetry install`.
    """
    content = """\
[[package]]
name = "demo"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false

[[package]]
name = "demo"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false

[package.source]
type = "git"
url = "https://github.com/demo/demo.git"
reference = "main"
resolved_reference = "123456"

[[package]]
name = "demo"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false

[package.source]
type = "directory"
url = "./folder"

[[package]]
name = "demo"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false

[package.source]
type = "file"
url = "./demo-1.0-cp39-win_amd64.whl"

[[package]]
name = "demo"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
develop = false

[package.source]
type = "url"
url = "https://example.com/demo-1.0-cp38-win_amd64.whl"

[metadata]
lock-version = "1.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"

[metadata.files]
# metadata.files are only tracked for non-direct origin and file dependencies
demo = [
    {file = "demo-1.0-cp39-win_amd64.whl", hash = "sha256"},
    {file = "demo-1.0.tar.gz", hash = "sha256"},
    {file = "demo-1.0-py3-none-any.whl", hash = "sha256"},
]
"""
    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    repository = locker.locked_repository()
    assert len(repository.packages) == 5
    assert {package.source_type for package in repository.packages} == {
        None,
        "git",
        "directory",
        "file",
        "url",
    }
    for package in repository.packages:
        if package.source_type is None:
            # non-direct origin package contains all files
            # with the current lockfile format we have no chance to determine
            # which files are correct, so we keep all for hash check
            # correct files are set later in Provider.complete_package()
            assert package.files == [
                {"file": "demo-1.0-cp39-win_amd64.whl", "hash": "sha256"},
                {"file": "demo-1.0.tar.gz", "hash": "sha256"},
                {"file": "demo-1.0-py3-none-any.whl", "hash": "sha256"},
            ]
        elif package.source_type == "file":
            assert package.files == [
                {"file": "demo-1.0-cp39-win_amd64.whl", "hash": "sha256"}
            ]
        else:
            package.files = []


def test_locker_dumps_packages_with_null_description(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_a = get_package("A", "1.0.0")
    package_a.description = None  # type: ignore[assignment]

    locker.set_lock_data(root, {package_a: transitive_info})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_does_not_dump_file_urls(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_a = get_package("A", "1.0")
    package_a.files = [
        {
            "file": "a-1.0.whl",
            "hash": "sha256:abcdef1234567890",
            "url": "https://example.org/a-1.0.whl",
        },
    ]

    locker.set_lock_data(root, {package_a: transitive_info})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = [
    {{file = "a-1.0.whl", hash = "sha256:abcdef1234567890"}},
]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_lock_file_should_not_have_mixed_types(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_a = get_package("A", "1.0.0")
    package_a.add_dependency(Factory.create_dependency("B", "^1.0.0"))
    package_a.add_dependency(
        Factory.create_dependency("B", {"version": ">=1.0.0", "optional": True})
    )
    package_a.requires[-1].activate()
    package_a.extras = {canonicalize_name("foo"): [get_dependency("B", ">=1.0.0")]}

    locker.set_lock_data(root, {package_a: transitive_info})

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
B = [
    {{version = "^1.0.0"}},
    {{version = ">=1.0.0", optional = true}},
]

[package.extras]
foo = ["B (>=1.0.0)"]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    assert content == expected


def test_reading_lock_file_should_raise_an_error_on_invalid_data(
    locker: Locker,
) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.extras]
foo = ["bar"]

[package.extras]
foo = ["bar"]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""
    with locker.lock.open("w", encoding="utf-8") as f:
        f.write(content)

    with pytest.raises(RuntimeError) as e:
        _ = locker.lock_data

    assert "Unable to read the lock file" in str(e.value)


def test_reading_lock_file_should_raise_an_error_on_missing_metadata(
    locker: Locker,
) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.source]
type = "legacy"
url = "https://foo.bar"
reference = "legacy"
"""
    with locker.lock.open("w", encoding="utf-8") as f:
        f.write(content)

    with pytest.raises(RuntimeError) as e:
        _ = locker.lock_data

    assert (
        "The lock file does not have a metadata entry.\nRegenerate the lock file with"
        " the `poetry lock` command." in str(e.value)
    )


def test_locking_legacy_repository_package_should_include_source_section(
    root: ProjectPackage, locker: Locker, transitive_info: TransitivePackageInfo
) -> None:
    package_a = Package(
        "A",
        "1.0.0",
        source_type="legacy",
        source_url="https://foo.bar",
        source_reference="legacy",
    )
    packages = {package_a: transitive_info}

    locker.set_lock_data(root, packages)

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.source]
type = "legacy"
url = "https://foo.bar"
reference = "legacy"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_should_emit_warnings_if_lock_version_is_newer_but_allowed(
    locker: Locker, caplog: LogCaptureFixture
) -> None:
    version = ".".join(Version.parse(Locker._VERSION).next_minor().text.split(".")[:2])
    content = f"""\
[metadata]
lock-version = "{version}"
python-versions = "~2.7 || ^3.4"
content-hash = "c3d07fca33fba542ef2b2a4d75bf5b48d892d21a830e2ad9c952ba5123a52f77"
"""
    caplog.set_level(logging.WARNING, logger="poetry.packages.locker")

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    _ = locker.lock_data

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.levelname == "WARNING"

    expected = """\
The lock file might not be compatible with the current version of Poetry.
Upgrade Poetry to ensure the lock file is read properly or, alternatively, \
regenerate the lock file with the `poetry lock` command.\
"""
    assert record.message == expected


def test_locker_should_raise_an_error_if_lock_version_is_newer_and_not_allowed(
    locker: Locker, caplog: LogCaptureFixture
) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[metadata]
lock-version = "3.0"
python-versions = "~2.7 || ^3.4"
content-hash = "c3d07fca33fba542ef2b2a4d75bf5b48d892d21a830e2ad9c952ba5123a52f77"
"""
    caplog.set_level(logging.WARNING, logger="poetry.packages.locker")

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    with pytest.raises(RuntimeError, match=r"^The lock file is not compatible"):
        _ = locker.lock_data


def test_locker_should_raise_an_error_if_no_lock_version(
    locker: Locker, caplog: LogCaptureFixture
) -> None:
    """Lock file prior Poetry 1.1 have no lock file version."""
    content = """\
[metadata]
python-versions = "~2.7 || ^3.4"
content-hash = "c3d07fca33fba542ef2b2a4d75bf5b48d892d21a830e2ad9c952ba5123a52f77"
"""
    caplog.set_level(logging.WARNING, logger="poetry.packages.locker")

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    with pytest.raises(RuntimeError, match=r"^The lock file is not compatible"):
        _ = locker.lock_data


def test_root_extras_dependencies_are_ordered(
    locker: Locker, root: ProjectPackage, fixture_base: Path
) -> None:
    Factory.create_dependency("B", "1.0.0", root_dir=fixture_base)
    Factory.create_dependency("C", "1.0.0", root_dir=fixture_base)
    package_first = Factory.create_dependency("first", "1.0.0", root_dir=fixture_base)
    package_second = Factory.create_dependency("second", "1.0.0", root_dir=fixture_base)
    package_third = Factory.create_dependency("third", "1.0.0", root_dir=fixture_base)

    root.extras = {
        canonicalize_name("C"): [package_third, package_second, package_first],
        canonicalize_name("B"): [package_first, package_second, package_third],
    }
    locker.set_lock_data(root, {})

    expected = f"""\
# {GENERATED_COMMENT}
package = []

[extras]
b = ["first", "second", "third"]
c = ["first", "second", "third"]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    assert content == expected


def test_extras_dependencies_are_ordered(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_a = get_package("A", "1.0.0")
    package_a.add_dependency(
        Factory.create_dependency(
            "B", {"version": "^1.0.0", "optional": True, "extras": ["c", "a", "b"]}
        )
    )
    package_a.requires[-1].activate()

    locker.set_lock_data(root, {package_a: transitive_info})

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
B = {{version = "^1.0.0", extras = ["a", "b", "c"], optional = true}}

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    assert content == expected


def test_locker_should_neither_emit_warnings_nor_raise_error_for_lower_compatible_versions(
    locker: Locker, caplog: LogCaptureFixture
) -> None:
    older_version = "1.1"
    content = f"""\
[metadata]
lock-version = "{older_version}"
python-versions = "~2.7 || ^3.4"
content-hash = "c3d07fca33fba542ef2b2a4d75bf5b48d892d21a830e2ad9c952ba5123a52f77"

[metadata.files]
"""
    caplog.set_level(logging.WARNING, logger="poetry.packages.locker")

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    _ = locker.lock_data

    assert len(caplog.records) == 0


def test_locker_dumps_groups_and_markers(
    locker: Locker,
    root: ProjectPackage,
    fixture_base: Path,
    transitive_info: TransitivePackageInfo,
) -> None:
    packages = {
        get_package("A", "1.0"): TransitivePackageInfo(
            0, {MAIN_GROUP}, {MAIN_GROUP: AnyMarker()}
        ),
        get_package("B", "1.0"): TransitivePackageInfo(
            0, {MAIN_GROUP}, {MAIN_GROUP: parse_marker('sys_platform == "win32"')}
        ),
        get_package("C", "1.0"): TransitivePackageInfo(
            0,
            {MAIN_GROUP, DEV_GROUP},
            {MAIN_GROUP: AnyMarker(), DEV_GROUP: AnyMarker()},
        ),
        get_package("D", "1.0"): TransitivePackageInfo(
            0,
            {MAIN_GROUP, DEV_GROUP},
            {
                MAIN_GROUP: parse_marker('sys_platform == "win32"'),
                DEV_GROUP: parse_marker('sys_platform == "win32"'),
            },
        ),
        get_package("E", "1.0"): TransitivePackageInfo(
            0,
            {MAIN_GROUP, DEV_GROUP},
            {
                MAIN_GROUP: parse_marker('sys_platform == "win32"'),
                DEV_GROUP: parse_marker('sys_platform == "linux"'),
            },
        ),
        get_package("F", "1.0"): TransitivePackageInfo(
            0,
            {MAIN_GROUP, DEV_GROUP},
            {
                MAIN_GROUP: parse_marker('sys_platform == "win32"'),
                DEV_GROUP: AnyMarker(),
            },
        ),
    }

    locker.set_lock_data(root, packages)

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[[package]]
name = "B"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
markers = "sys_platform == \\"win32\\""
files = []

[[package]]
name = "C"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main", "dev"]
files = []

[[package]]
name = "D"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main", "dev"]
markers = "sys_platform == \\"win32\\""
files = []

[[package]]
name = "E"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main", "dev"]
files = []
markers = {{main = "sys_platform == \\"win32\\"", dev = "sys_platform == \\"linux\\""}}

[[package]]
name = "F"
version = "1.0"
description = ""
optional = false
python-versions = "*"
groups = ["main", "dev"]
files = []
markers = {{main = "sys_platform == \\"win32\\""}}

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_dumps_dependency_information_correctly(
    locker: Locker,
    root: ProjectPackage,
    fixture_base: Path,
    transitive_info: TransitivePackageInfo,
) -> None:
    package_a = get_package("A", "1.0.0")
    package_a.add_dependency(
        Factory.create_dependency(
            "B", {"path": "project_with_extras", "develop": True}, root_dir=fixture_base
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "C",
            {"path": "directory/project_with_transitive_directory_dependencies"},
            root_dir=fixture_base,
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "D", {"path": "distributions/demo-0.1.0.tar.gz"}, root_dir=fixture_base
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "E", {"url": "https://files.pythonhosted.org/poetry-1.2.0.tar.gz"}
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "F", {"git": "https://github.com/python-poetry/poetry.git", "branch": "foo"}
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "G",
            {
                "git": "https://github.com/python-poetry/poetry.git",
                "subdirectory": "bar",
            },
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "H", {"git": "https://github.com/python-poetry/poetry.git", "tag": "baz"}
        )
    )
    package_a.add_dependency(
        Factory.create_dependency(
            "I", {"git": "https://github.com/python-poetry/poetry.git", "rev": "spam"}
        )
    )

    packages = {package_a: transitive_info}

    locker.set_lock_data(root, packages)

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.dependencies]
B = {{path = "project_with_extras", develop = true}}
C = {{path = "directory/project_with_transitive_directory_dependencies"}}
D = {{path = "distributions/demo-0.1.0.tar.gz"}}
E = {{url = "https://files.pythonhosted.org/poetry-1.2.0.tar.gz"}}
F = {{git = "https://github.com/python-poetry/poetry.git", branch = "foo"}}
G = {{git = "https://github.com/python-poetry/poetry.git", subdirectory = "bar"}}
H = {{git = "https://github.com/python-poetry/poetry.git", tag = "baz"}}
I = {{git = "https://github.com/python-poetry/poetry.git", rev = "spam"}}

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_dumps_subdir(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    package_git_with_subdirectory = Package(
        "git-package-subdir",
        "1.2.3",
        source_type="git",
        source_url="https://github.com/python-poetry/poetry.git",
        source_reference="develop",
        source_resolved_reference="123456",
        source_subdirectory="subdir",
    )

    locker.set_lock_data(root, {package_git_with_subdirectory: transitive_info})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "git-package-subdir"
version = "1.2.3"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []
develop = false

[package.source]
type = "git"
url = "https://github.com/python-poetry/poetry.git"
reference = "develop"
resolved_reference = "123456"
subdirectory = "subdir"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_dumps_dependency_extras_in_correct_order(
    locker: Locker,
    root: ProjectPackage,
    fixture_base: Path,
    transitive_info: TransitivePackageInfo,
) -> None:
    package_a = get_package("A", "1.0.0")
    Factory.create_dependency("B", "1.0.0", root_dir=fixture_base)
    Factory.create_dependency("C", "1.0.0", root_dir=fixture_base)
    package_first = Factory.create_dependency("first", "1.0.0", root_dir=fixture_base)
    package_second = Factory.create_dependency("second", "1.0.0", root_dir=fixture_base)
    package_third = Factory.create_dependency("third", "1.0.0", root_dir=fixture_base)

    package_a.extras = {
        canonicalize_name("C"): [package_third, package_second, package_first],
        canonicalize_name("B"): [package_first, package_second, package_third],
    }

    locker.set_lock_data(root, {package_a: transitive_info})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.extras]
b = ["first (==1.0.0)", "second (==1.0.0)", "third (==1.0.0)"]
c = ["first (==1.0.0)", "second (==1.0.0)", "third (==1.0.0)"]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_dumps_extras_with_constraints(
    locker: Locker,
    root: ProjectPackage,
    fixture_base: Path,
    transitive_info: TransitivePackageInfo,
) -> None:
    package_a = get_package("A", "1.0.0")

    package_httpx_new = Factory.create_dependency(
        "httpx",
        {"version": "2.0.0", "python": ">=3.7", "extras": ["brotli"]},
        root_dir=fixture_base,
    )

    package_httpx_old = Factory.create_dependency(
        "httpx", {"version": "1.0.0", "python": "<3.7"}, root_dir=fixture_base
    )

    package_a.extras = {
        canonicalize_name("http"): [package_httpx_new, package_httpx_old],
    }

    locker.set_lock_data(root, {package_a: transitive_info})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.extras]
http = ["httpx (==1.0.0) ; python_version < \\"3.7\\"", "httpx[brotli] (==2.0.0) ; python_version >= \\"3.7\\""]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    assert content == expected


def test_locker_properly_loads_extras_with_constraints(locker: Locker) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "A"
version = "1.0.0"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.extras]
http = ["httpx (==1.0.0) ; python_version < \\"3.7\\"", "httpx[brotli] (==2.0.0) ; python_version >= \\"3.7\\""]

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    packages = locker.locked_repository().packages

    assert len(packages) == 1

    package = packages[0]
    assert len(package.extras) == 1

    httpx_deps = package.extras[canonicalize_name("http")]
    assert len(httpx_deps) == 2

    assert httpx_deps[0].constraint == Version.parse("1.0.0")
    assert str(httpx_deps[0].python_constraint) == "<3.7"

    assert httpx_deps[1].constraint == Version.parse("2.0.0")
    assert str(httpx_deps[1].python_constraint) == ">=3.7"
    assert httpx_deps[1].extras == {"brotli"}


def test_locked_repository_uses_root_dir_of_package(
    locker: Locker, mocker: MockerFixture
) -> None:
    content = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "lib-a"
version = "0.1.0"
description = ""
optional = false
python-versions = "^2.7.9"
groups = ["main"]
develop = true
file = []

[package.dependencies]
lib-b = {{path = "../libB", develop = true}}

[package.source]
type = "directory"
url = "lib/libA"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(content)

    create_dependency_patch = mocker.patch(
        "poetry.factory.Factory.create_dependency", autospec=True
    )
    locker.locked_repository()

    create_dependency_patch.assert_called_once_with(
        "lib-b", {"develop": True, "path": "../libB"}, root_dir=mocker.ANY
    )
    call_kwargs = create_dependency_patch.call_args[1]
    root_dir = call_kwargs["root_dir"]
    assert root_dir.match("*/lib/libA")
    # relative_to raises an exception if not relative - is_relative_to comes in py3.9
    assert root_dir.relative_to(locker.lock.parent.resolve()) is not None


@pytest.mark.parametrize(
    ("local_config", "legacy"),
    [
        ({}, True),
        ({"dependencies": [uuid.uuid4().hex]}, True),
        ({"dependencies": [uuid.uuid4().hex], "source": [uuid.uuid4().hex]}, True),
        ({"dependencies": [uuid.uuid4().hex], "extras": [uuid.uuid4().hex]}, True),
        (
            {
                "dependencies": [uuid.uuid4().hex],
                "dev-dependencies": [uuid.uuid4().hex],
            },
            True,
        ),
        (
            {
                "dependencies": [uuid.uuid4().hex],
                "dev-dependencies": None,
            },
            True,
        ),
        ({"dependencies": [uuid.uuid4().hex], "group": [uuid.uuid4().hex]}, False),
    ],
)
def test_content_hash_with_legacy_is_compatible(
    local_config: dict[str, list[str]], legacy: bool, locker: Locker
) -> None:
    """Legacy generation if there is no group."""

    def _get_legacy_content_hash() -> str:
        relevant_content = {}
        for key in locker._legacy_keys:
            relevant_content[key] = local_config.get(key)

        content_hash = sha256(
            json.dumps(relevant_content, sort_keys=True).encode()
        ).hexdigest()

        return content_hash

    locker = locker.__class__(
        lock=locker.lock,
        pyproject_data={"tool": {"poetry": local_config}},
    )

    old_content_hash = _get_legacy_content_hash()
    content_hash = locker._get_content_hash()

    if legacy:
        assert content_hash == old_content_hash
    else:
        assert content_hash != old_content_hash


@pytest.mark.parametrize(
    ("project", "legacy"),
    [
        ({"name": "foo"}, True),  # irrelevant key
        ({"requires-python": ">=3.9"}, False),
        ({"dependencies": ["bar"]}, False),
        ({"dependencies": []}, False),  # relevant even if empty
        ({"optional-dependencies": "..."}, False),
    ],
)
@pytest.mark.parametrize(
    "local_config",
    [
        {},  # empty
        {"dependencies": [uuid.uuid4().hex]},  # only legacy
        {
            "dependencies": [uuid.uuid4().hex],
            "group": [uuid.uuid4().hex],
        },  # legacy and new
    ],
)
def test_content_hash_with_project_section(
    project: dict[str, Any],
    local_config: dict[str, list[str]],
    legacy: bool,
    locker: Locker,
) -> None:
    """Legacy generation if there is no project section."""

    def _get_legacy_content_hash() -> str:
        relevant_content = {}
        for key in locker._relevant_keys:
            data = local_config.get(key)

            if data is None and key not in locker._legacy_keys:
                continue

            relevant_content[key] = data

        return sha256(json.dumps(relevant_content, sort_keys=True).encode()).hexdigest()

    locker = locker.__class__(
        lock=locker.lock,
        pyproject_data={"project": project, "tool": {"poetry": local_config}},
    )

    old_content_hash = _get_legacy_content_hash()
    content_hash = locker._get_content_hash()

    if legacy:
        assert content_hash == old_content_hash
    else:
        assert content_hash != old_content_hash


@pytest.mark.parametrize(
    ("groups", "legacy"),
    [
        ({}, True),
        ({"foo": []}, False),
    ],
)
@pytest.mark.parametrize(
    "project",
    [
        {"name": "foo"},  # irrelevant key
        {"requires-python": ">=3.9"},  # relevant key
    ],
)
@pytest.mark.parametrize(
    "local_config",
    [
        {},  # empty
        {"dependencies": [uuid.uuid4().hex]},  # only legacy
        {
            "dependencies": [uuid.uuid4().hex],
            "group": [uuid.uuid4().hex],
        },  # legacy and new
    ],
)
def test_content_hash_with_dependency_groups_section(
    groups: dict[str, Any],
    project: dict[str, Any],
    local_config: dict[str, Any],
    legacy: bool,
    locker: Locker,
) -> None:
    """Legacy generation if there is no dependency-groups section."""

    def _get_legacy_content_hash() -> str:
        project_content = project
        tool_poetry_content = local_config

        relevant_project_content = {}
        for key in locker._relevant_project_keys:
            data = project_content.get(key)
            if data is not None:
                relevant_project_content[key] = data

        relevant_poetry_content: dict[str, Any] = {}
        for key in locker._relevant_keys:
            data = tool_poetry_content.get(key)

            if data is None and (
                # Special handling for legacy keys is just for backwards compatibility,
                # and thereby not required if there is relevant content in [project].
                key not in locker._legacy_keys or relevant_project_content
            ):
                continue

            relevant_poetry_content[key] = data

        if relevant_project_content:
            relevant_content = {
                "project": relevant_project_content,
                "tool": {"poetry": relevant_poetry_content},
            }
        else:
            # For backwards compatibility, we have to put the relevant content
            # of the [tool.poetry] section at top level!
            relevant_content = relevant_poetry_content

        return sha256(json.dumps(relevant_content, sort_keys=True).encode()).hexdigest()

    locker = locker.__class__(
        lock=locker.lock,
        pyproject_data={
            "project": project,
            "dependency-groups": groups,
            "tool": {"poetry": local_config},
        },
    )

    old_content_hash = _get_legacy_content_hash()
    content_hash = locker._get_content_hash()

    if legacy:
        assert content_hash == old_content_hash
    else:
        assert content_hash != old_content_hash


def test_lock_file_resolves_file_url_symlinks(
    root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    """
    Create directories and file structure as follows:

    d1/
    d1/testsymlink -> d1/d2/d3
    d1/d2/d3/lock_file
    d1/d4/source_file

    Using the testsymlink as the Locker.lock file path should correctly resolve to
    the real physical path of the source_file when calculating the relative path
    from the lock_file, i.e. "../../d4/source_file" instead of the unresolved path
    from the symlink itself which would have been "../d4/source_file"

    See https://github.com/python-poetry/poetry/issues/5849
    """
    with tempfile.TemporaryDirectory() as d1:
        symlink_path = Path(d1).joinpath("testsymlink")
        with (
            tempfile.TemporaryDirectory(dir=d1) as d2,
            tempfile.TemporaryDirectory(dir=d1) as d4,
            tempfile.TemporaryDirectory(dir=d2) as d3,
            tempfile.NamedTemporaryFile(dir=d4) as source_file,
            tempfile.NamedTemporaryFile(dir=d3) as lock_file,
        ):
            lock_file.close()
            try:
                os.symlink(Path(d3), symlink_path)
            except OSError:
                if sys.platform == "win32":
                    # os.symlink requires either administrative privileges or developer
                    # mode on Win10, throwing an OSError if neither is active.
                    # Test is not possible in that case.
                    return
                raise
            locker = Locker(symlink_path / lock_file.name, {})

            package_local = Package(
                "local-package",
                "1.2.3",
                source_type="file",
                source_url=source_file.name,
                source_reference="develop",
                source_resolved_reference="123456",
            )
            packages = {
                package_local: transitive_info,
            }

            locker.set_lock_data(root, packages)

            with locker.lock.open(encoding="utf-8") as f:
                content = f.read()

            expected = f"""\
# {GENERATED_COMMENT}

[[package]]
name = "local-package"
version = "1.2.3"
description = ""
optional = false
python-versions = "*"
groups = ["main"]
files = []

[package.source]
type = "file"
url = "{
                Path(
                    os.path.relpath(
                        Path(source_file.name).resolve().as_posix(),
                        Path(Path(lock_file.name).parent).resolve().as_posix(),
                    )
                ).as_posix()
            }"
reference = "develop"
resolved_reference = "123456"

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

            assert content == expected


def test_lockfile_is_not_rewritten_if_only_poetry_version_changed(
    locker: Locker, root: ProjectPackage
) -> None:
    generated_comment_old_version = GENERATED_COMMENT.replace(__version__, "1.3.2")
    assert generated_comment_old_version != GENERATED_COMMENT
    old_content = f"""\
# {generated_comment_old_version}
package = []

[metadata]
lock-version = "2.1"
python-versions = "*"
content-hash = "115cf985d932e9bf5f540555bbdd75decbb62cac81e399375fc19f6277f8c1d8"
"""

    with open(locker.lock, "w", encoding="utf-8") as f:
        f.write(old_content)

    assert not locker.set_lock_data(root, {})

    with locker.lock.open(encoding="utf-8") as f:
        content = f.read()

    assert content == old_content


def test_lockfile_keep_eol(
    locker: Locker, root: ProjectPackage, transitive_info: TransitivePackageInfo
) -> None:
    sep = "\n" if os.linesep == "\r\n" else "\r\n"

    with open(locker.lock, "wb") as f:
        f.write((sep * 10).encode())

    packages = {Package("test", version="0.0.1"): transitive_info}

    assert locker.set_lock_data(root, packages)

    with locker.lock.open(encoding="utf-8", newline="") as f:
        line, *_ = f.read().splitlines(keepends=True)

    if sep == "\r\n":
        assert line.endswith("\r\n")
    else:
        assert not line.endswith("\r\n")



from __future__ import annotations

import pytest

from packaging.utils import canonicalize_name
from poetry.core.constraints.version.version import Version

from poetry.repositories.link_sources.json import SimpleJsonPage
from poetry.repositories.link_sources.json import SimpleRepositoryJsonRootPage


@pytest.fixture
def root_page() -> SimpleRepositoryJsonRootPage:
    names = ["poetry", "poetry-core", "requests"]

    return SimpleRepositoryJsonRootPage(
        {
            "meta": {"api-version": "1.4"},
            "projects": [{"name": name} for name in names],
        }
    )


def test_root_page_package_names(root_page: SimpleRepositoryJsonRootPage) -> None:
    assert root_page.package_names == ["poetry", "poetry-core", "requests"]


def test_attributes() -> None:
    content = {
        "files": [
            # minimal
            {"url": "https://example.org/demo-0.1.whl"},
            # all (with non-default values)
            {
                "url": "https://example.org/demo-0.1.tar.gz",
                "requires-python": ">=3.6",
                "yanked": True,
                "hashes": {"sha256": "abcd1234"},
                "core-metadata": True,
            },
        ]
    }
    page = SimpleJsonPage("https://example.org", content)

    assert page.url == "https://example.org"
    links = list(page.links)
    assert len(links) == 2

    assert links[0].url == "https://example.org/demo-0.1.whl"
    assert links[0].requires_python is None
    assert links[0].yanked is False
    assert links[0].hashes == {}
    assert links[0].has_metadata is False

    assert links[1].url == "https://example.org/demo-0.1.tar.gz"
    assert links[1].requires_python == ">=3.6"
    assert links[1].yanked is True
    assert links[1].hashes == {"sha256": "abcd1234"}
    assert links[1].has_metadata is True


@pytest.mark.parametrize(
    ("yanked", "expected"),
    [
        ((None, None), False),
        ((False, False), False),
        ((True, False), False),
        ((False, True), False),
        ((True, True), True),
        (("reason", True), "reason"),
        ((True, "reason"), "reason"),
        (("reason", "reason"), "reason"),
        (("reason 1", "reason 2"), "reason 1\nreason 2"),
    ],
)
def test_yanked(
    yanked: tuple[str | None, str | None],
    expected: bool | str,
) -> None:
    content = {
        "files": [
            {"url": "https://example.org/demo-0.1.tar.gz", "yanked": yanked[0]},
            {"url": "https://example.org/demo-0.1.whl", "yanked": yanked[1]},
        ]
    }
    if yanked[0] is None:
        del content["files"][0]["yanked"]
    if yanked[1] is None:
        del content["files"][1]["yanked"]
    page = SimpleJsonPage("https://example.org", content)

    assert page.yanked(canonicalize_name("demo"), Version.parse("0.1")) == expected


@pytest.mark.parametrize(
    ("metadata", "expected_has_metadata", "expected_metadata_hashes"),
    [
        ({}, False, {}),
        # new
        ({"core-metadata": False}, False, {}),
        ({"core-metadata": True}, True, {}),
        (
            {"core-metadata": {"sha1": "1234", "sha256": "abcd"}},
            True,
            {"sha1": "1234", "sha256": "abcd"},
        ),
        ({"core-metadata": {}}, False, {}),
        (
            {"core-metadata": {"sha1": "1234", "sha256": "abcd"}},
            True,
            {"sha1": "1234", "sha256": "abcd"},
        ),
        # old
        ({"dist-info-metadata": False}, False, {}),
        ({"dist-info-metadata": True}, True, {}),
        ({"dist-info-metadata": {"sha256": "abcd"}}, True, {"sha256": "abcd"}),
        ({"dist-info-metadata": {}}, False, {}),
        (
            {"dist-info-metadata": {"sha1": "1234", "sha256": "abcd"}},
            True,
            {"sha1": "1234", "sha256": "abcd"},
        ),
        # conflicting (new wins)
        ({"core-metadata": False, "dist-info-metadata": True}, False, {}),
        (
            {"core-metadata": False, "dist-info-metadata": {"sha256": "abcd"}},
            False,
            {},
        ),
        ({"core-metadata": True, "dist-info-metadata": False}, True, {}),
        (
            {"core-metadata": True, "dist-info-metadata": {"sha256": "abcd"}},
            True,
            {},
        ),
        (
            {"core-metadata": {"sha256": "abcd"}, "dist-info-metadata": False},
            True,
            {"sha256": "abcd"},
        ),
        (
            {"core-metadata": {"sha256": "abcd"}, "dist-info-metadata": True},
            True,
            {"sha256": "abcd"},
        ),
        (
            {
                "core-metadata": {"sha256": "abcd"},
                "dist-info-metadata": {"sha256": "1234"},
            },
            True,
            {"sha256": "abcd"},
        ),
    ],
)
def test_metadata(
    metadata: dict[str, bool | dict[str, str]],
    expected_has_metadata: bool,
    expected_metadata_hashes: dict[str, str],
) -> None:
    content = {"files": [{"url": "https://example.org/demo-0.1.whl", **metadata}]}
    page = SimpleJsonPage("https://example.org", content)

    link = next(page.links)
    assert link.has_metadata is expected_has_metadata
    assert link.metadata_hashes == expected_metadata_hashes


@pytest.mark.parametrize(
    ("url", "repo_url", "expected"),
    (
        (
            "https://example.org/files/demo-0.1.whl",
            "https://example.org/simple/",
            "https://example.org/files/demo-0.1.whl",
        ),
        (
            "demo-0.1.whl",
            "https://example.org/simple/",
            "https://example.org/simple/demo-0.1.whl",
        ),
    ),
)
def test_base_url(url: str, repo_url: str, expected: str) -> None:
    page = SimpleJsonPage(repo_url, {"files": [{"url": url}]})
    link = next(iter(page.links))
    assert link.url == expected



from __future__ import annotations

import json
import re

from typing import TYPE_CHECKING
from typing import Any
from urllib.parse import urlparse

import pytest
import responses

from packaging.utils import parse_sdist_filename
from packaging.utils import parse_wheel_filename

from poetry.repositories.pypi_repository import PyPiRepository
from tests.helpers import FIXTURE_PATH_DISTRIBUTIONS
from tests.helpers import FIXTURE_PATH_REPOSITORIES_PYPI


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from requests import PreparedRequest

    from tests.types import HttpRequestCallback
    from tests.types import HttpResponse
    from tests.types import PackageDistributionLookup


pytest_plugins = [
    "tests.repositories.fixtures.legacy",
    "tests.repositories.fixtures.python_hosted",
]


@pytest.fixture
def package_distribution_locations() -> list[Path]:
    return [
        FIXTURE_PATH_REPOSITORIES_PYPI / "dists",
        FIXTURE_PATH_REPOSITORIES_PYPI / "dists" / "mocked",
        FIXTURE_PATH_REPOSITORIES_PYPI / "stubbed",
        FIXTURE_PATH_DISTRIBUTIONS,
    ]


@pytest.fixture
def package_json_locations() -> list[Path]:
    return [
        FIXTURE_PATH_REPOSITORIES_PYPI / "json",
        FIXTURE_PATH_REPOSITORIES_PYPI / "json" / "mocked",
    ]


@pytest.fixture
def package_metadata_locations() -> list[Path]:
    return [
        FIXTURE_PATH_REPOSITORIES_PYPI / "metadata",
        FIXTURE_PATH_REPOSITORIES_PYPI / "metadata" / "mocked",
    ]


@pytest.fixture
def package_distribution_lookup(
    package_distribution_locations: list[Path],
) -> PackageDistributionLookup:
    def lookup(name: str) -> Path | None:
        for location in package_distribution_locations:
            fixture = location / name
            if fixture.exists():
                return fixture
        return None

    return lookup


@pytest.fixture
def with_disallowed_pypi_search_html(
    http: responses.RequestsMock, pypi_repository: PyPiRepository
) -> None:
    def search_callback(request: PreparedRequest) -> HttpResponse:
        search_html = FIXTURE_PATH_REPOSITORIES_PYPI.joinpath(
            "search", "search-disallowed.html"
        )
        return 200, {}, search_html.read_bytes()

    search_url_regex = re.compile(r"https://pypi\.org/search(\?(.*))?$")
    http.remove(responses.GET, search_url_regex)
    http.add_callback(
        responses.GET,
        search_url_regex,
        callback=search_callback,
    )


@pytest.fixture(autouse=True)
def pypi_repository(
    http: responses.RequestsMock,
    legacy_repository_html_callback: HttpRequestCallback,
    package_json_locations: list[Path],
    mock_files_python_hosted: None,
) -> PyPiRepository:
    def default_callback(request: PreparedRequest) -> HttpResponse:
        return 404, {}, b"Not Found"

    def search_callback(request: PreparedRequest) -> HttpResponse:
        search_html = FIXTURE_PATH_REPOSITORIES_PYPI.joinpath("search", "search.html")
        return 200, {}, search_html.read_bytes()

    def simple_callback(request: PreparedRequest) -> HttpResponse:
        if request.headers.get("Accept") == "application/vnd.pypi.simple.v1+json":
            return json_callback(request)
        return legacy_repository_html_callback(request)

    def _get_json_filepath(name: str, version: str | None = None) -> Path | None:
        for base in package_json_locations:
            if not version:
                fixture = base / f"{name}.json"
            else:
                fixture = base / name / f"{version}.json"

            if fixture.exists():
                return fixture

        return None

    def json_callback(request: PreparedRequest) -> HttpResponse:
        assert request.url
        path = urlparse(request.url).path
        parts = path.rstrip("/").split("/")[2:]
        name = parts[0]
        version = parts[1] if len(parts) == 3 else None
        fixture = _get_json_filepath(name, version)

        if fixture is None or not fixture.exists():
            return default_callback(request)

        return 200, {}, fixture.read_bytes()

    http.add_callback(
        responses.GET,
        re.compile(r"https://pypi\.org/search(\?(.*))?$"),
        callback=search_callback,
    )

    http.add_callback(
        responses.GET,
        re.compile(r"https://pypi\.org/pypi/(.*)?/json$"),
        callback=json_callback,
    )

    http.add_callback(
        responses.GET,
        re.compile(r"https://pypi\.org/pypi/(?!.*?/json$)(.*)$"),
        callback=default_callback,
    )

    http.add_callback(
        responses.GET,
        re.compile(r"https://pypi\.org/simple/?(.*)?$"),
        callback=simple_callback,
    )

    return PyPiRepository(disable_cache=True, fallback=False)


@pytest.fixture
def get_pypi_file_info(
    package_json_locations: list[Path],
) -> Callable[[str], dict[str, Any]]:
    def get_file_info(name: str) -> dict[str, Any]:
        if name.endswith(".whl"):
            package_name, version, _build, _tags = parse_wheel_filename(name)
        else:
            package_name, version = parse_sdist_filename(name)
        path = package_json_locations[0] / package_name
        if not path.exists():
            raise RuntimeError(
                f"Fixture for {package_name} not found in pypi.org json fixtures"
            )
        path /= f"{version}.json"
        if not path.exists():
            raise RuntimeError(
                f"Fixture for {package_name} {version} not found in pypi.org json fixtures"
            )
        with path.open("rb") as f:
            content = json.load(f)
        for url in content["urls"]:
            if url["filename"] == name:
                return url  # type: ignore[no-any-return]
        raise RuntimeError(f"No URL in pypi.org json fixture of {name} {version}")

    return get_file_info



from __future__ import annotations

import uuid

from typing import TYPE_CHECKING

import pytest

from poetry.utils.cache import FileCache


if TYPE_CHECKING:
    from pathlib import Path

    from tests.conftest import Config


@pytest.fixture
def repository_cache_dir(config: Config) -> Path:
    return config.repository_cache_directory


@pytest.fixture
def repositories() -> list[str]:
    return [f"01_{uuid.uuid4()}", f"02_{uuid.uuid4()}"]


@pytest.fixture
def repository_dirs(
    repository_cache_dir: Path,
    repositories: list[str],
) -> list[Path]:
    return [
        repository_cache_dir / repositories[0],
        repository_cache_dir / repositories[1],
    ]


@pytest.fixture
def caches(
    repository_dirs: list[Path],
) -> list[FileCache[dict[str, str]]]:
    repository_dirs[0].mkdir(parents=True)
    repository_dirs[1].mkdir(parents=True)

    caches: list[FileCache[dict[str, str]]] = [
        FileCache(path=repository_dirs[0]),
        FileCache(path=repository_dirs[1]),
    ]

    caches[0].remember(
        "cachy:0.1", lambda: {"name": "cachy", "version": "0.1"}, minutes=None
    )
    caches[0].remember(
        "cleo:0.2", lambda: {"name": "cleo", "version": "0.2"}, minutes=None
    )

    caches[1].remember(
        "cachy:0.1", lambda: {"name": "cachy", "version": "0.1"}, minutes=None
    )
    # different version of same package, other entry than in first cache
    caches[1].remember(
        "cashy:0.2", lambda: {"name": "cashy", "version": "0.2"}, minutes=None
    )

    return caches



from __future__ import annotations

from poetry.core.packages.package import Package

from poetry.utils.env.mock_env import MockEnv
from poetry.utils.log_utils import format_build_wheel_log


def test_format_build_wheel_log() -> None:
    env = MockEnv(version_info=(3, 13, 1), platform="win32", platform_machine="AMD64")
    package = Package(name="demo", version="1.2.3")
    result = format_build_wheel_log(package, env)
    expected = (
        " <info>Building a wheel file for demo for Python 3.13.1 on win32-AMD64</info>"
    )
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import pytest
import tomlkit

from poetry.core.packages.dependency_group import DependencyGroup
from poetry.core.packages.package import Package

from poetry.factory import Factory
from tests.helpers import TestLocker
from tests.helpers import get_package


if TYPE_CHECKING:
    from collections.abc import Callable

    from cleo.testers.command_tester import CommandTester
    from pytest_mock import MockerFixture
    from tomlkit import TOMLDocument

    from poetry.poetry import Poetry
    from poetry.repositories import Repository
    from tests.helpers import PoetryTestApplication
    from tests.helpers import TestRepository
    from tests.types import CommandTesterFactory
    from tests.types import FixtureDirGetter
    from tests.types import ProjectFactory


@pytest.fixture
def poetry_with_up_to_date_lockfile(
    project_factory: ProjectFactory, fixture_dir: FixtureDirGetter
) -> Callable[[str], Poetry]:
    def get_poetry(fixture_name: str) -> Poetry:
        source = fixture_dir(fixture_name)

        poetry = project_factory(
            name="foobar",
            pyproject_content=(source / "pyproject.toml").read_text(encoding="utf-8"),
            poetry_lock_content=(source / "poetry.lock").read_text(encoding="utf-8"),
        )

        assert isinstance(poetry.locker, TestLocker)
        poetry.locker.locked(True)
        return poetry

    return get_poetry


@pytest.fixture()
def tester(command_tester_factory: CommandTesterFactory) -> CommandTester:
    return command_tester_factory("remove")


def test_remove_from_project_and_poetry(
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("bar", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()

    project_dependencies: dict[str, Any] = tomlkit.parse(
        """\
[project]
dependencies = [
    "foo>=2.0",
    "bar>=1.0",
]
"""
    )

    poetry_dependencies: dict[str, Any] = tomlkit.parse(
        """\
[tool.poetry.dependencies]
foo = "^2.0.0"
bar = "^1.0.0"

"""
    )

    pyproject["project"]["dependencies"] = project_dependencies["project"][
        "dependencies"
    ]
    pyproject["tool"]["poetry"]["dependencies"] = poetry_dependencies["tool"]["poetry"][
        "dependencies"
    ]
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo", "^2.0.0"))
    app.poetry.package.add_dependency(Factory.create_dependency("bar", "^1.0.0"))

    tester.execute("foo")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    project_dependencies = pyproject["project"]["dependencies"]
    assert "foo>=2.0" not in project_dependencies
    assert "bar>=1.0" in project_dependencies
    poetry_dependencies = pyproject["tool"]["poetry"]["dependencies"]
    assert "foo" not in poetry_dependencies
    assert "bar" in poetry_dependencies

    expected_project_string = """\
dependencies = [
    "bar>=1.0",
]
"""
    expected_poetry_string = """\

[tool.poetry.dependencies]
bar = "^1.0.0"

"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected_project_string = expected_project_string.replace("\n", "\r\n")
        expected_poetry_string = expected_poetry_string.replace("\n", "\r\n")

    assert expected_project_string in string_content
    assert expected_poetry_string in string_content


def test_remove_from_pep735_group_and_poetry_group(
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("bar", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()

    pep735_dependencies: dict[str, Any] = tomlkit.parse(
        """\
[dependency-groups]
dev = [
    "foo>=2.0",
    "bar>=1.0",
]
"""
    )

    poetry_dependencies: dict[str, Any] = tomlkit.parse(
        """\
[tool.poetry.group.dev.dependencies]
foo = "^2.0.0"
bar = "^1.0.0"

"""
    )

    pyproject["dependency-groups"] = pep735_dependencies["dependency-groups"]
    pyproject["tool"]["poetry"]["group"] = poetry_dependencies["tool"]["poetry"][
        "group"
    ]
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["dev"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("bar", "^1.0.0", groups=["dev"])
    )

    tester.execute("foo")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    pep735_dependencies = pyproject["dependency-groups"]["dev"]
    assert "foo>=2.0" not in pep735_dependencies
    assert "bar>=1.0" in pep735_dependencies
    poetry_dependencies = pyproject["tool"]["poetry"]["group"]["dev"]["dependencies"]
    assert "foo" not in poetry_dependencies
    assert "bar" in poetry_dependencies

    expected_pep735_string = """\
[dependency-groups]
dev = [
    "bar>=1.0",
]
"""
    expected_poetry_string = """\

[tool.poetry.group.dev.dependencies]
bar = "^1.0.0"

"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected_pep735_string = expected_pep735_string.replace("\n", "\r\n")
        expected_poetry_string = expected_poetry_string.replace("\n", "\r\n")

    assert expected_pep735_string in string_content
    assert expected_poetry_string in string_content


@pytest.mark.parametrize("pep_735", [True, False])
def test_remove_without_specific_group_removes_from_all_groups(
    pep_735: bool,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    Removing without specifying a group removes packages from all groups.
    """
    installed.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    pyproject["tool"]["poetry"]["dependencies"]["foo"] = "^2.0.0"

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
    "baz (>=1.0,<2.0)",
]
"""
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]

    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"
baz = "^1.0.0"

"""
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"]["group"] = groups_content["tool"]["poetry"]["group"]

    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo", "^2.0.0"))
    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["bar"])
    )

    tester.execute("foo")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]
    assert "foo" not in content["dependencies"]

    if pep_735:
        assert not any("foo" in dep for dep in pyproject["dependency-groups"]["bar"])
        assert any("baz" in dep for dep in pyproject["dependency-groups"]["bar"])
        expected = """\
[dependency-groups]
bar = [
    "baz (>=1.0,<2.0)",
]
"""
    else:
        assert "foo" not in content["group"]["bar"]["dependencies"]
        assert "baz" in content["group"]["bar"]["dependencies"]

        expected = """\
[tool.poetry.group.bar.dependencies]
baz = "^1.0.0"
"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected = expected.replace("\n", "\r\n")

    assert expected in string_content


@pytest.mark.parametrize("pep_735", [True, False])
def test_remove_with_specific_group_removes_from_specific_groups(
    pep_735: bool,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    Removing with a specific group given removes packages only from this group.
    """
    installed.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    pyproject["tool"]["poetry"]["dependencies"]["foo"] = "^2.0.0"

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
    "baz (>=1.0,<2.0)",
]
    """
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]

    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"
baz = "^1.0.0"
    """
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"]["group"] = groups_content["tool"]["poetry"]["group"]

    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo", "^2.0.0"))
    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["bar"])
    )

    tester.execute("foo --group bar")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]
    assert "foo" in content["dependencies"]

    if pep_735:
        assert not any("foo" in dep for dep in pyproject["dependency-groups"]["bar"])
        assert any("baz" in dep for dep in pyproject["dependency-groups"]["bar"])
        expected = """\
[dependency-groups]
bar = [
    "baz (>=1.0,<2.0)",
]
"""
    else:
        assert "foo" not in content["group"]["bar"]["dependencies"]
        assert "baz" in content["group"]["bar"]["dependencies"]

        expected = """\
[tool.poetry.group.bar.dependencies]
baz = "^1.0.0"
"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected = expected.replace("\n", "\r\n")

    assert expected in string_content


@pytest.mark.parametrize("pep_735", [True, False])
def test_remove_does_not_keep_empty_groups(
    pep_735: bool,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    Empty groups are automatically discarded after package removal.
    """
    installed.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    pyproject["tool"]["poetry"]["dependencies"]["foo"] = "^2.0.0"

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
    "baz (>=1.0,<2.0)",
]
    """
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]
    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"
baz = "^1.0.0"

"""
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"]["group"] = groups_content["tool"]["poetry"]["group"]
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo", "^2.0.0"))
    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["bar"])
    )

    tester.execute("foo baz --group bar")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]

    assert "foo" in content["dependencies"]

    if pep_735:
        assert "bar" not in pyproject.get("dependency-groups", {})
        assert "dependency-groups" not in pyproject
    else:
        # The group 'bar' should be removed entirely from the configuration
        assert "group" not in content
        content = cast("TOMLDocument", content)
        assert "[tool.poetry.group.bar]" not in content.as_string()
        assert "[tool.poetry.group]" not in content.as_string()


@pytest.mark.parametrize("pep_735", [True, False])
def test_remove_canonicalized_named_removes_dependency_correctly(
    pep_735: bool,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    Removing a dependency using a canonicalized named removes the dependency.
    """
    installed.add_package(Package("foo-bar", "2.0.0"))
    repo.add_package(Package("foo-bar", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    pyproject["tool"]["poetry"]["dependencies"]["foo-bar"] = "^2.0.0"

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo-bar (>=2.0,<3.0)",
    "baz (>=1.0,<2.0)",
]
"""
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]
    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo-bar = "^2.0.0"
baz = "^1.0.0"

"""
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"].value._insert_after(
            "dependencies", "group", groups_content["tool"]["poetry"]["group"]
        )
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo-bar", "^2.0.0"))
    app.poetry.package.add_dependency(
        Factory.create_dependency("foo-bar", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["bar"])
    )

    tester.execute("Foo_Bar")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]

    assert "foo-bar" not in content["dependencies"]

    if pep_735:
        assert not any("foo" in dep for dep in pyproject["dependency-groups"]["bar"])
        assert any("baz" in dep for dep in pyproject["dependency-groups"]["bar"])
        expected = """\
[dependency-groups]
bar = [
    "baz (>=1.0,<2.0)",
]
"""
    else:
        assert "foo-bar" not in content["group"]["bar"]["dependencies"]
        assert "baz" in content["group"]["bar"]["dependencies"]

        expected = """\
[tool.poetry.group.bar.dependencies]
baz = "^1.0.0"
"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected = expected.replace("\n", "\r\n")

    assert expected in string_content


def test_remove_package_does_not_exist(
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    command_tester_factory: CommandTesterFactory,
) -> None:
    repo.add_package(Package("foo", "2.0.0"))

    original_content = app.poetry.file.read().as_string()

    with pytest.raises(ValueError) as e:
        tester.execute("foo")

    assert str(e.value) == "The following packages were not found: foo"
    assert app.poetry.file.read().as_string() == original_content


def test_remove_package_no_dependencies(
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    command_tester_factory: CommandTesterFactory,
) -> None:
    repo.add_package(Package("foo", "2.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    assert "dependencies" not in pyproject["project"]
    del pyproject["tool"]["poetry"]["dependencies"]
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)
    app.poetry.package._dependency_groups = {}

    with pytest.raises(ValueError) as e:
        tester.execute("foo")

    assert str(e.value) == "The following packages were not found: foo"


def test_remove_command_should_not_write_changes_upon_installer_errors(
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    command_tester_factory: CommandTesterFactory,
    mocker: MockerFixture,
) -> None:
    repo.add_package(Package("foo", "2.0.0"))

    command_tester_factory("add").execute("foo")

    mocker.patch("poetry.installation.installer.Installer.run", return_value=1)

    original_content = app.poetry.file.read().as_string()

    tester.execute("foo")

    assert app.poetry.file.read().as_string() == original_content


@pytest.mark.parametrize(
    "fixture_name", ["up_to_date_lock", "up_to_date_lock_non_package"]
)
def test_remove_with_dry_run_keep_files_intact(
    fixture_name: str,
    poetry_with_up_to_date_lockfile: Callable[[str], Poetry],
    repo: TestRepository,
    command_tester_factory: CommandTesterFactory,
) -> None:
    poetry = poetry_with_up_to_date_lockfile(fixture_name)
    tester = command_tester_factory("remove", poetry=poetry)

    original_pyproject_content = poetry.file.read()
    original_lockfile_content = poetry._locker.lock_data

    repo.add_package(get_package("docker", "4.3.1"))

    tester.execute("docker --dry-run")

    assert poetry.file.read() == original_pyproject_content
    assert poetry._locker.lock_data == original_lockfile_content


@pytest.mark.parametrize(
    "fixture_name", ["up_to_date_lock", "up_to_date_lock_non_package"]
)
def test_remove_performs_uninstall_op(
    fixture_name: str,
    poetry_with_up_to_date_lockfile: Callable[[str], Poetry],
    command_tester_factory: CommandTesterFactory,
    installed: Repository,
) -> None:
    installed.add_package(get_package("docker", "4.3.1"))
    poetry = poetry_with_up_to_date_lockfile(fixture_name)
    tester = command_tester_factory("remove", poetry=poetry)

    tester.execute("docker")

    expected = """\
Updating dependencies
Resolving dependencies...

Package operations: 0 installs, 0 updates, 1 removal

  - Removing docker (4.3.1)

Writing lock file
"""

    assert tester.io.fetch_output() == expected


@pytest.mark.parametrize(
    "fixture_name", ["up_to_date_lock", "up_to_date_lock_non_package"]
)
def test_remove_with_lock_does_not_perform_uninstall_op(
    fixture_name: str,
    poetry_with_up_to_date_lockfile: Callable[[str], Poetry],
    command_tester_factory: CommandTesterFactory,
    installed: Repository,
) -> None:
    installed.add_package(get_package("docker", "4.3.1"))
    poetry = poetry_with_up_to_date_lockfile(fixture_name)
    tester = command_tester_factory("remove", poetry=poetry)

    tester.execute("docker --lock")

    expected = """\
Updating dependencies
Resolving dependencies...

Writing lock file
"""

    assert tester.io.fetch_output() == expected


@pytest.mark.parametrize("pep_735", [True, False])
def test_remove_from_nested_pep735_group_and_poetry_group(
    pep_735: bool,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    Test removing packages from nested dependency groups with `include-group`(pep735) or `include-groups`(poetry).
    """
    installed.add_package(Package("foo", "2.0.0"))
    installed.add_package(Package("baz", "1.0.0"))
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()
    pyproject["tool"]["poetry"]["dependencies"]["foo"] = "^2.0.0"

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
]
foobar = [
    { include-group = "bar" },
    "baz (>=1.0,<2.0)",
]
"""
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]

    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"

[tool.poetry.group.foobar]
include-groups = [
    "bar",
]

[tool.poetry.group.foobar.dependencies]
baz = "^1.0.0"
"""
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"]["group"] = groups_content["tool"]["poetry"]["group"]
    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(Factory.create_dependency("foo", "^2.0.0"))
    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["foobar"])
    )

    tester.execute("baz")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]

    if pep_735:
        assert "baz" not in pyproject["dependency-groups"]["foobar"]
        assert any(
            isinstance(item, dict) and item.get("include-group") == "bar"
            for item in pyproject["dependency-groups"]["foobar"]
        )
        expected = """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
]
foobar = [
    { include-group = "bar" },
]
"""
    else:
        # This also checks that the include-groups section is not removed even if its explicit dependencies are empty
        assert "dependencies" not in content["group"]["foobar"]
        expected = """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"

[tool.poetry.group.foobar]
include-groups = [
    "bar",
]
"""
    pyproject = cast("TOMLDocument", pyproject)
    string_content = pyproject.as_string()
    if "\r\n" in string_content:
        # consistent line endings
        expected = expected.replace("\n", "\r\n")

    assert expected in string_content


@pytest.mark.parametrize("pep_735", [True, False])
@pytest.mark.parametrize("args", ["", " --group bar"])
def test_remove_group_cleans_up_include_group_references(
    pep_735: bool,
    args: str,
    tester: CommandTester,
    app: PoetryTestApplication,
    repo: TestRepository,
    installed: Repository,
) -> None:
    """
    When a group is removed, any `include-group` references to it in other
    groups should also be cleaned up.
    """
    installed.add_package(Package("foo", "2.0.0"))
    installed.add_package(Package("baz", "1.0.0"))
    repo.add_package(Package("foo", "2.0.0"))
    repo.add_package(Package("baz", "1.0.0"))

    pyproject: dict[str, Any] = app.poetry.file.read()

    if pep_735:
        groups_content: dict[str, Any] = tomlkit.parse(
            """\
[dependency-groups]
bar = [
    "foo (>=2.0,<3.0)",
]
foobar = [
    { include-group = "bar" },
]
foobar2 = [
    "baz (>=1.0)",
    { include-group = "bar" },
    "baz (<=3.0)",
]
foobar3 = [
    { include-group = "bar" },
    { include-group = "foobar2" },
]
"""
        )
        pyproject["dependency-groups"] = groups_content["dependency-groups"]
    else:
        groups_content = tomlkit.parse(
            """\
[tool.poetry.group.bar.dependencies]
foo = "^2.0.0"

[tool.poetry.group.foobar]
include-groups = [
    "bar",
]

[tool.poetry.group.foobar2]
include-groups = [
    "bar",
]

[tool.poetry.group.foobar2.dependencies]
baz = "(>=1.0,<=3.0)"

[tool.poetry.group.foobar3]
include-groups = [
    "bar",
    "foobar2",
]
"""
        )
        groups_content = cast("dict[str, Any]", groups_content)
        pyproject["tool"]["poetry"]["group"] = groups_content["tool"]["poetry"]["group"]

    pyproject = cast("TOMLDocument", pyproject)
    app.poetry.file.write(pyproject)

    app.poetry.package.add_dependency(
        Factory.create_dependency("foo", "^2.0.0", groups=["bar"])
    )
    app.poetry.package.add_dependency(
        Factory.create_dependency("baz", "^1.0.0", groups=["foobar2"])
    )
    foobar = DependencyGroup("foobar")
    foobar.include_dependency_group(app.poetry.package.dependency_group("bar"))
    app.poetry.package.add_dependency_group(foobar)
    foobar2 = DependencyGroup("foobar2")
    foobar2.include_dependency_group(app.poetry.package.dependency_group("bar"))
    app.poetry.package.add_dependency_group(foobar2)
    foobar3 = DependencyGroup("foobar3")
    foobar3.include_dependency_group(app.poetry.package.dependency_group("bar"))
    foobar3.include_dependency_group(app.poetry.package.dependency_group("foobar2"))
    app.poetry.package.add_dependency_group(foobar3)

    # Remove all packages from the "bar" group, which should delete the group
    # and also clean up references to it in "foobar"
    tester.execute(f"foo{args}")

    pyproject = app.poetry.file.read()
    pyproject = cast("dict[str, Any]", pyproject)
    content = pyproject["tool"]["poetry"]

    if pep_735:
        # "bar" group should be removed
        assert "bar" not in pyproject.get("dependency-groups", {})
        # "foobar" group should also be removed since it only had include-group
        assert "foobar" not in pyproject.get("dependency-groups", {})
        # "foobar2" should have its include-group cleaned up
        assert "foobar2" in pyproject.get("dependency-groups", {})
        assert pyproject["dependency-groups"]["foobar2"] == [
            "baz (>=1.0)",
            "baz (<=3.0)",
        ]
        # "foobar3" should have its include-groups cleaned up
        assert "foobar3" in pyproject.get("dependency-groups", {})
        assert pyproject["dependency-groups"]["foobar3"] == [
            {"include-group": "foobar2"}
        ]
    else:
        # "bar" group should be removed
        assert "bar" not in content.get("group", {})
        # "foobar" group should also be removed since it only had include-group
        assert "foobar" not in content.get("group", {})
        # "foobar2" should have its include-groups cleaned up
        assert "foobar2" in content.get("group", {})
        assert "include-groups" not in content["group"]["foobar2"]
        assert "dependencies" in content["group"]["foobar2"]
        assert content["group"]["foobar2"]["dependencies"] == {"baz": "(>=1.0,<=3.0)"}
        # "foobar3" should have its include-groups cleaned up
        assert "foobar3" in content.get("group", {})
        assert "include-groups" in content["group"]["foobar3"]
        assert content["group"]["foobar3"]["include-groups"] == ["foobar2"]



from __future__ import annotations

import os

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import pytest
import tomlkit

from poetry.core.constraints.version import Version

from poetry.console.commands.env.use import EnvUseCommand
from poetry.toml.file import TOMLFile
from poetry.utils.env import MockEnv
from poetry.utils.env.python.exceptions import NoCompatiblePythonVersionFoundError
from tests.console.commands.env.helpers import build_venv
from tests.console.commands.env.helpers import check_output_wrapper


if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from cleo.testers.command_tester import CommandTester
    from pytest_mock import MockerFixture

    from poetry.utils.env.base_env import PythonVersion
    from tests.types import CommandTesterFactory
    from tests.types import MockedPythonRegister


@pytest.fixture(autouse=True)
def setup(mocker: MockerFixture) -> None:
    if "VIRTUAL_ENV" in os.environ:
        del os.environ["VIRTUAL_ENV"]


@pytest.fixture(autouse=True)
def mock_subprocess_calls(
    setup: None, current_python: PythonVersion, mocker: MockerFixture
) -> None:
    mocker.patch(
        "subprocess.check_output",
        side_effect=check_output_wrapper(Version.from_parts(*current_python[:3])),
    )
    mocker.patch(
        "subprocess.Popen.communicate",
        side_effect=[("/prefix", None), ("/prefix", None), ("/prefix", None)],
    )


@pytest.fixture
def tester(command_tester_factory: CommandTesterFactory) -> CommandTester:
    return command_tester_factory("env use")


def test_activate_activates_non_existing_virtualenv_no_envs_file(
    mocker: MockerFixture,
    tester: CommandTester,
    venv_cache: Path,
    venv_name: str,
    venvs_in_cache_config: None,
    mocked_python_register: MockedPythonRegister,
    with_no_active_python: MagicMock,
) -> None:
    mocked_python_register("3.7.1")
    mock_build_env = mocker.patch(
        "poetry.utils.env.EnvManager.build_venv", side_effect=build_venv
    )
    envs_file = TOMLFile(venv_cache / "envs.toml")

    assert not envs_file.exists()
    assert not list(venv_cache.iterdir())

    tester.execute("3.7")

    venv_py37 = venv_cache / f"{venv_name}-py3.7"
    mock_build_env.assert_called_with(
        venv_py37,
        executable=Path("/usr/bin/python3.7"),
        flags={
            "always-copy": False,
            "system-site-packages": False,
            "no-pip": False,
        },
        prompt="simple-project-py3.7",
    )

    assert envs_file.exists()
    envs: dict[str, Any] = envs_file.read()
    assert envs[venv_name]["minor"] == "3.7"
    assert envs[venv_name]["patch"] == "3.7.1"

    assert (
        tester.io.fetch_error()
        == f"Creating virtualenv {venv_py37.name} in {venv_py37.parent}\n"
    )
    assert tester.io.fetch_output() == f"Using virtualenv: {venv_py37}\n"


@pytest.mark.parametrize("use_poetry_python", [True, False])
def test_activate_does_not_activate_non_existing_virtualenv_with_unsupported_version(
    tester: CommandTester,
    venv_cache: Path,
    venv_name: str,
    venvs_in_cache_config: None,
    mocked_python_register: MockedPythonRegister,
    with_no_active_python: MagicMock,
    use_poetry_python: bool,
) -> None:
    mocked_python_register("3.7.1")
    mocked_python_register("3.8.2")
    command = tester.command
    assert isinstance(command, EnvUseCommand)
    command.poetry.package.python_versions = "~3.8"
    command.poetry.config.merge(
        {"virtualenvs": {"use-poetry-python": use_poetry_python}}
    )

    assert not list(venv_cache.iterdir())

    with pytest.raises(NoCompatiblePythonVersionFoundError):
        tester.execute("3.7")

    assert not list(venv_cache.iterdir())


def test_get_prefers_explicitly_activated_virtualenvs_over_env_var(
    tester: CommandTester,
    current_python: PythonVersion,
    venv_cache: Path,
    venv_name: str,
    venvs_in_cache_config: None,
    mocked_python_register: MockedPythonRegister,
) -> None:
    os.environ["VIRTUAL_ENV"] = "/environment/prefix"

    python_minor = ".".join(str(v) for v in current_python[:2])
    python_patch = ".".join(str(v) for v in current_python[:3])
    venv_dir = venv_cache / f"{venv_name}-py{python_minor}"
    venv_dir.mkdir(parents=True, exist_ok=True)

    envs_file = TOMLFile(venv_cache / "envs.toml")
    doc = tomlkit.document()
    doc[venv_name] = {"minor": python_minor, "patch": python_patch}
    envs_file.write(doc)

    mocked_python_register(python_patch)

    tester.execute(python_minor)

    expected = f"""\
Using virtualenv: {venv_dir}
"""

    assert tester.io.fetch_output() == expected


def test_get_prefers_explicitly_activated_non_existing_virtualenvs_over_env_var(
    mocker: MockerFixture,
    tester: CommandTester,
    current_python: PythonVersion,
    venv_cache: Path,
    venv_name: str,
    venvs_in_cache_config: None,
    mocked_python_register: MockedPythonRegister,
) -> None:
    os.environ["VIRTUAL_ENV"] = "/environment/prefix"

    python_minor = ".".join(str(v) for v in current_python[:2])
    venv_dir = venv_cache / f"{venv_name}-py{python_minor}"

    mocked_python_register(python_minor)

    mocker.patch(
        "poetry.utils.env.EnvManager._env",
        new_callable=mocker.PropertyMock,
        return_value=MockEnv(
            path=Path("/environment/prefix"),
            base=Path("/base/prefix"),
            version_info=current_python,
            is_venv=True,
        ),
    )
    mocker.patch("poetry.utils.env.EnvManager.build_venv", side_effect=build_venv)

    tester.execute(python_minor)

    assert (
        tester.io.fetch_error()
        == f"Creating virtualenv {venv_dir.name} in {venv_dir.parent}\n"
    )
    assert tester.io.fetch_output() == f"Using virtualenv: {venv_dir}\n"



from __future__ import annotations

from typing import Any

import pytest

from packaging.utils import NormalizedName
from packaging.utils import canonicalize_name
from poetry.core.constraints.version import Version

from poetry.inspection.info import PackageInfo
from poetry.repositories.cached_repository import CachedRepository


class MockCachedRepository(CachedRepository):
    def _get_release_info(
        self, name: NormalizedName, version: Version
    ) -> dict[str, Any]:
        raise NotImplementedError


@pytest.fixture
def release_info() -> PackageInfo:
    return PackageInfo(
        name="mylib",
        version="1.0",
        summary="",
        requires_dist=[],
        requires_python=">=3.9",
        files=[
            {
                "file": "mylib-1.0-py3-none-any.whl",
                "hash": "sha256:dummyhashvalue1234567890abcdef",
            },
            {
                "file": "mylib-1.0.tar.gz",
                "hash": "sha256:anotherdummyhashvalueabcdef1234567890",
            },
        ],
        cache_version=str(CachedRepository.CACHE_VERSION),
    )


@pytest.fixture
def outdated_release_info() -> PackageInfo:
    return PackageInfo(
        name="mylib",
        version="1.0",
        summary="",
        requires_dist=[],
        requires_python=">=3.9",
        files=[
            {
                "file": "mylib-1.0-py3-none-any.whl",
                "hash": "sha256:dummyhashvalue1234567890abcdef",
            }
        ],
        cache_version=str(CachedRepository.CACHE_VERSION),
    )


@pytest.mark.parametrize("disable_cache", [False, True])
def test_get_release_info_cache(
    release_info: PackageInfo, outdated_release_info: PackageInfo, disable_cache: bool
) -> None:
    repo = MockCachedRepository("mock", disable_cache=disable_cache)
    repo._get_release_info = lambda name, version: outdated_release_info.asdict()  # type: ignore[method-assign]

    name = canonicalize_name("mylib")
    version = Version.parse("1.0")
    assert len(repo.get_release_info(name, version).files) == 1

    # without disable_cache: cached value is returned even if the underlying data has changed
    # with disable_cache: cached value is ignored and updated data is returned
    repo._get_release_info = lambda name, version: release_info.asdict()  # type: ignore[method-assign]
    assert len(repo.get_release_info(name, version).files) == (
        2 if disable_cache else 1
    )

    # after clearing the cache entry, updated data is returned
    repo.forget(name, version)
    assert len(repo.get_release_info(name, version).files) == 2



from __future__ import annotations

import re

from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import pytest
import responses


if TYPE_CHECKING:
    from collections.abc import Iterator

    from requests import PreparedRequest

    from tests.types import HttpResponse
    from tests.types import PythonHostedFileMocker


@pytest.fixture
def mock_files_python_hosted_factory(
    http: responses.RequestsMock,
) -> PythonHostedFileMocker:
    def factory(
        distribution_locations: list[Path], metadata_locations: list[Path]
    ) -> None:
        def file_callback(request: PreparedRequest) -> HttpResponse:
            assert request.url
            name = Path(urlparse(request.url).path).name

            locations = (
                metadata_locations
                if name.endswith(".metadata")
                else distribution_locations
            )

            for location in locations:
                fixture = location / name
                if fixture.exists():
                    return 200, {}, fixture.read_bytes()

            return 404, {}, b"Not Found"

        def mock_file_callback(request: PreparedRequest) -> HttpResponse:
            return 200, {}, b""

        http.add_callback(
            responses.GET,
            re.compile(r"^https://files\.pythonhosted\.org/.*$"),
            callback=file_callback,
        )

        http.add_callback(
            responses.GET,
            re.compile(r"^https://mock\.pythonhosted\.org/.*$"),
            callback=mock_file_callback,
        )

    return factory


@pytest.fixture
def mock_files_python_hosted(
    mock_files_python_hosted_factory: PythonHostedFileMocker,
    package_distribution_locations: list[Path],
    package_metadata_locations: list[Path],
) -> Iterator[None]:
    mock_files_python_hosted_factory(
        distribution_locations=package_distribution_locations,
        metadata_locations=package_metadata_locations,
    )
    yield None



from __future__ import annotations

import json

from typing import TYPE_CHECKING

import pytest
import tomlkit

from poetry.__version__ import __version__
from poetry.console.commands.self.self_command import SelfCommand


if TYPE_CHECKING:
    from cleo.testers.command_tester import CommandTester

    from tests.types import CommandTesterFactory


@pytest.fixture()
def tester(command_tester_factory: CommandTesterFactory) -> CommandTester:
    return command_tester_factory("self show")


@pytest.mark.parametrize("options", ["", "--format json", "--format text"])
def test_show_format(tester: CommandTester, options: str) -> None:
    pyproject_content = {
        "tool": {
            "poetry": {
                "name": "poetry-instance",
                "version": __version__,
                "dependencies": {"python": "^3.9", "poetry": __version__},
            }
        }
    }
    lock_content = {
        "package": [
            {
                "name": "poetry",
                "version": __version__,
                "optional": False,
                "platform": "*",
                "python-versions": "*",
                "files": [],
            },
        ],
        "metadata": {
            "lock-version": "2.0",
            "python-versions": "^3.9",
            "content-hash": "123456789",
        },
    }
    if "json" in options:
        expected = json.dumps(
            [
                {
                    "name": "poetry",
                    "installed_status": "installed",
                    "version": __version__,
                    "description": "",
                }
            ]
        )
    else:
        expected = f"poetry {__version__}"
    system_pyproject_file = SelfCommand.get_default_system_pyproject_file()
    system_pyproject_file.write_text(tomlkit.dumps(pyproject_content), encoding="utf-8")
    system_pyproject_file.parent.joinpath("poetry.lock").write_text(
        tomlkit.dumps(lock_content), encoding="utf-8"
    )
    assert tester.execute(options) == 0
    assert tester.io.fetch_output().strip() == expected



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeVar

import pytest

from cleo.testers.application_tester import ApplicationTester

from poetry.console.application import Application


if TYPE_CHECKING:
    from pathlib import Path

    from poetry.utils.cache import FileCache

T = TypeVar("T")


@pytest.fixture
def tester() -> ApplicationTester:
    app = Application()

    tester = ApplicationTester(app)
    return tester


@pytest.mark.parametrize("inputs", ["yes", "no"])
def test_cache_clear_all(
    tester: ApplicationTester,
    repository_cache_dir: Path,
    repositories: list[str],
    repository_dirs: list[Path],
    caches: list[FileCache[dict[str, str]]],
    inputs: str,
) -> None:
    exit_code = tester.execute("cache clear --all", inputs=inputs)

    assert exit_code == 0
    assert tester.io.fetch_output() == ""

    if inputs == "yes":
        assert not repository_dirs[0].exists() or not any(repository_dirs[0].iterdir())
        assert not repository_dirs[1].exists() or not any(repository_dirs[1].iterdir())
        assert not caches[0].has("cachy:0.1")
        assert not caches[0].has("cleo:0.2")
        assert not caches[1].has("cachy:0.1")
        assert not caches[1].has("cashy:0.2")
    else:
        assert any((repository_cache_dir / repositories[0]).iterdir())
        assert any((repository_cache_dir / repositories[1]).iterdir())
        assert caches[0].has("cachy:0.1")
        assert caches[0].has("cleo:0.2")
        assert caches[1].has("cachy:0.1")
        assert caches[1].has("cashy:0.2")


@pytest.mark.parametrize("inputs", ["yes", "no"])
def test_cache_clear_all_one_cache(
    tester: ApplicationTester,
    repository_cache_dir: Path,
    repositories: list[str],
    repository_dirs: list[Path],
    caches: list[FileCache[dict[str, str]]],
    inputs: str,
) -> None:
    exit_code = tester.execute(f"cache clear {repositories[0]} --all", inputs=inputs)

    assert exit_code == 0
    assert tester.io.fetch_output() == ""

    if inputs == "yes":
        assert not repository_dirs[0].exists() or not any(repository_dirs[0].iterdir())
        assert not caches[0].has("cachy:0.1")
        assert not caches[0].has("cleo:0.2")
    else:
        assert any((repository_cache_dir / repositories[0]).iterdir())
        assert caches[0].has("cachy:0.1")
        assert caches[0].has("cleo:0.2")

    assert any((repository_cache_dir / repositories[1]).iterdir())
    assert caches[1].has("cachy:0.1")
    assert caches[1].has("cashy:0.2")


def test_cache_clear_all_no_entries(tester: ApplicationTester) -> None:
    exit_code = tester.execute("cache clear --all")

    assert exit_code == 0
    assert tester.io.fetch_output().strip() == "No cache entries"


def test_cache_clear_all_one_cache_no_entries(
    tester: ApplicationTester,
    repository_cache_dir: Path,
    repositories: list[str],
) -> None:
    exit_code = tester.execute(f"cache clear {repositories[0]} --all")

    assert exit_code == 0

    assert tester.io.fetch_output().strip() == f"No cache entries for {repositories[0]}"


@pytest.mark.parametrize("with_repo", [False, True])
def test_cache_clear_missing_option(
    tester: ApplicationTester, repositories: list[str], with_repo: bool
) -> None:
    command = f"cache clear {repositories[0]}" if with_repo else "cache clear"
    exit_code = tester.execute(command)

    assert exit_code == 1
    assert (
        "Add the --all option if you want to clear all cache entries"
        in tester.io.fetch_error()
    )


@pytest.mark.parametrize("inputs", ["yes", "no"])
@pytest.mark.parametrize("package_name", ["cachy", "Cachy"])
def test_cache_clear_pkg(
    tester: ApplicationTester,
    repositories: list[str],
    caches: list[FileCache[dict[str, str]]],
    package_name: str,
    inputs: str,
) -> None:
    exit_code = tester.execute(
        f"cache clear {repositories[1]}:{package_name}:0.1", inputs=inputs
    )

    assert exit_code == 0
    assert tester.io.fetch_output() == ""

    if inputs == "yes":
        assert not caches[1].has("cachy:0.1")
        assert caches[1].has("cashy:0.2")
    else:
        assert caches[1].has("cachy:0.1")
        assert caches[1].has("cashy:0.2")

    assert caches[0].has("cachy:0.1")



from __future__ import annotations

import shutil

from typing import TYPE_CHECKING
from typing import cast

import pytest

from dulwich.client import FetchPackResult
from dulwich.repo import Repo

from poetry.console.exceptions import PoetryRuntimeError
from poetry.vcs.git.backend import Git
from poetry.vcs.git.backend import GitRefSpec
from poetry.vcs.git.backend import is_revision_sha
from poetry.vcs.git.backend import peeled_tag
from poetry.vcs.git.backend import urlpathjoin
from tests.helpers import MOCK_DEFAULT_GIT_REVISION


if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture

    from tests.vcs.git.git_fixture import TempRepoFixture


VALID_SHA = "c5c7624ef64f34d9f50c3b7e8118f7f652fddbbd"

FULL_SHA_MAIN = "f7c3bc1d808e04732adf679965ccc34ca7ae3441"
FULL_SHA_TAG = "d4f6c2a8b9e1073451f28c96a5db7e3f9c2a8b7e"
SHORT_SHA = "f7c3bc1d"


@pytest.fixture()
def repo_mock(mocker: MockerFixture) -> Repo:
    repo = mocker.MagicMock(spec=Repo)

    repo.get_config.return_value.get.return_value = (
        b"https://github.com/python-poetry/poetry.git"
    )

    repo.head.return_value = MOCK_DEFAULT_GIT_REVISION.encode("utf-8")

    # Mock object store for short SHA resolution
    repo.object_store = mocker.MagicMock()
    repo.object_store.iter_prefix.return_value = [FULL_SHA_MAIN.encode()]

    return cast("Repo", repo)


@pytest.fixture()
def fetch_pack_result(mocker: MockerFixture) -> FetchPackResult:
    mock_fetch_pack_result = mocker.MagicMock(spec=FetchPackResult)
    mock_fetch_pack_result.refs = {
        b"refs/heads/main": FULL_SHA_MAIN.encode(),
        b"refs/heads/feature": b"a9b8c7d6e5f4321098765432109876543210abcd",
        b"refs/tags/v1.0.0": FULL_SHA_TAG.encode(),
        peeled_tag(b"refs/tags/v1.0.0"): FULL_SHA_TAG.encode(),
        b"HEAD": FULL_SHA_MAIN.encode(),
    }
    mock_fetch_pack_result.symrefs = {b"HEAD": b"refs/heads/main"}

    return cast("FetchPackResult", mock_fetch_pack_result)


def test_invalid_revision_sha() -> None:
    result = is_revision_sha("invalid_input")
    assert result is False


def test_valid_revision_sha() -> None:
    result = is_revision_sha(VALID_SHA)
    assert result is True


def test_invalid_revision_sha_min_len() -> None:
    result = is_revision_sha("c5c7")
    assert result is False


def test_invalid_revision_sha_max_len() -> None:
    result = is_revision_sha(VALID_SHA + "42")
    assert result is False


@pytest.mark.parametrize(
    ("url"),
    [
        "git@github.com:python-poetry/poetry.git",
        "https://github.com/python-poetry/poetry.git",
        "https://github.com/python-poetry/poetry",
        "https://github.com/python-poetry/poetry/",
    ],
)
def test_get_name_from_source_url(url: str) -> None:
    name = Git.get_name_from_source_url(url)
    assert name == "poetry"


@pytest.mark.parametrize(("tag"), ["my-tag", b"my-tag"])
def test_peeled_tag(tag: str | bytes) -> None:
    tag = peeled_tag("my-tag")
    assert tag == b"my-tag^{}"


def test_get_remote_url(repo_mock: Repo) -> None:
    assert (
        Git.get_remote_url(repo_mock) == "https://github.com/python-poetry/poetry.git"
    )


def test_get_revision(repo_mock: Repo) -> None:
    assert Git.get_revision(repo_mock) == MOCK_DEFAULT_GIT_REVISION


def test_info(repo_mock: Repo) -> None:
    info = Git.info(repo_mock)

    assert info.origin == "https://github.com/python-poetry/poetry.git"
    assert (
        info.revision == MOCK_DEFAULT_GIT_REVISION
    )  # revision already mocked in helper


@pytest.mark.parametrize(
    "url, expected_result",
    [
        ("ssh://git@github.com/org/repo", "ssh://git@github.com/other-repo"),
        ("ssh://git@github.com/org/repo/", "ssh://git@github.com/org/other-repo"),
    ],
)
def test_urlpathjoin(url: str, expected_result: str) -> None:
    path = "../other-repo"
    result = urlpathjoin(url, path)
    assert result == expected_result


def test_git_refspec() -> None:
    git_ref = GitRefSpec("main", "1234", "v2")

    assert git_ref.branch == "main"
    assert git_ref.revision == "1234"
    assert git_ref.tag == "v2"
    assert git_ref.ref == b"HEAD"


@pytest.mark.parametrize(
    "refspec, expected_ref, expected_branch, expected_revision, expected_tag",
    [
        # Basic parameter tests
        (
            GitRefSpec(branch="main"),
            b"refs/heads/main",
            "main",
            None,
            None,
        ),
        (
            GitRefSpec(tag="v1.0.0"),
            peeled_tag(b"refs/tags/v1.0.0"),
            None,
            None,
            "v1.0.0",
        ),
        (
            GitRefSpec(branch="refs/heads/feature"),
            b"refs/heads/feature",
            "refs/heads/feature",
            None,
            None,
        ),
        # Cross-parameter resolution tests
        (
            GitRefSpec(revision="v1.0.0"),
            peeled_tag(b"refs/tags/v1.0.0"),
            None,
            None,
            "v1.0.0",
        ),
        (
            GitRefSpec(revision="main"),
            b"refs/heads/main",
            "main",
            None,
            None,
        ),
        (
            GitRefSpec(branch="v1.0.0"),
            peeled_tag(b"refs/tags/v1.0.0"),
            None,
            None,
            "v1.0.0",
        ),
        (
            GitRefSpec(revision="refs/heads/main"),
            b"refs/heads/main",
            "refs/heads/main",
            None,
            None,
        ),
        # SHA resolution tests with realistic values
        (
            GitRefSpec(revision=SHORT_SHA),
            b"refs/heads/main",
            None,
            FULL_SHA_MAIN,
            None,
        ),
        (
            GitRefSpec(revision=FULL_SHA_MAIN),
            b"refs/heads/main",
            None,
            FULL_SHA_MAIN,
            None,
        ),
    ],
)
def test_git_ref_spec_resolve(
    fetch_pack_result: FetchPackResult,
    repo_mock: Repo,
    refspec: GitRefSpec,
    expected_ref: bytes,
    expected_branch: str | None,
    expected_revision: str | None,
    expected_tag: str | None,
) -> None:
    refspec.resolve(fetch_pack_result, repo_mock)

    assert refspec.ref == expected_ref
    assert refspec.branch == expected_branch
    assert refspec.revision == expected_revision
    assert refspec.tag == expected_tag


@pytest.mark.skip_git_mock
def test_clone_success(tmp_path: Path, temp_repo: TempRepoFixture) -> None:
    source_root_dir = tmp_path / "test-repo"
    Git.clone(
        url=temp_repo.path.as_uri(), source_root=source_root_dir, name="clone-test"
    )

    target_dir = source_root_dir / "clone-test"
    assert (target_dir / ".git").is_dir()


@pytest.mark.skip_git_mock
def test_short_sha_not_in_head(tmp_path: Path, temp_repo: TempRepoFixture) -> None:
    source_root_dir = tmp_path / "test-repo"
    Git.clone(
        url=temp_repo.path.as_uri(),
        revision=temp_repo.middle_commit[:6],
        name="clone-test",
        source_root=source_root_dir,
    )

    target_dir = source_root_dir / "clone-test"
    assert (target_dir / ".git").is_dir()


@pytest.mark.skip_git_mock
def test_clone_existing_locked_tag(tmp_path: Path, temp_repo: TempRepoFixture) -> None:
    source_root_dir = tmp_path / "test-repo"
    source_url = temp_repo.path.as_uri()
    Git.clone(url=source_url, source_root=source_root_dir, name="clone-test")

    tag_ref = source_root_dir / "clone-test" / ".git" / "refs" / "tags" / "v1"
    assert tag_ref.is_file()

    tag_ref_lock = tag_ref.with_name("v1.lock")
    shutil.copy(tag_ref, tag_ref_lock)

    with pytest.raises(PoetryRuntimeError) as exc_info:
        Git.clone(url=source_url, source_root=source_root_dir, name="clone-test")

    expected_short = (
        f"Failed to clone {source_url} at 'refs/heads/main',"
        f" unable to acquire file lock for {tag_ref}."
    )
    assert str(exc_info.value) == expected_short
    assert exc_info.value.get_text(debug=True, strip=True) == (
        f"{expected_short}\n\n"
        "Note: This error arises from interacting with the specified vcs source"
        " and is likely not a Poetry issue.\n"
        "This issue could be caused by any of the following;\n\n"
        "- another process is holding the file lock\n"
        "- another process crashed while holding the file lock\n\n"
        f"Try again later or remove the {tag_ref_lock} manually"
        " if you are sure no other process is holding it."
    )



from __future__ import annotations

from subprocess import CalledProcessError
from typing import TYPE_CHECKING
from typing import Literal
from typing import cast

import pytest

from poetry.core.constraints.version import Version

from poetry.console.exceptions import PoetryRuntimeError
from poetry.utils.env.python.installer import PythonDownloadNotFoundError
from poetry.utils.env.python.installer import PythonInstallationError
from poetry.utils.env.python.installer import PythonInstaller


if TYPE_CHECKING:
    from pathlib import Path
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture


@pytest.fixture(autouse=True)
def mock_get_download_link(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "pbs_installer.get_download_link",
        return_value=(mocker.Mock(major=3, minor=9, micro=1), None),
    )


def test_python_installer_version() -> None:
    installer = PythonInstaller(request="3.9.1")
    assert installer.version == Version.from_parts(3, 9, 1)


def test_python_installer_version_not_found(mock_get_download_link: MagicMock) -> None:
    mock_get_download_link.return_value = []
    installer = PythonInstaller(request="3.9.1")
    with pytest.raises(PythonDownloadNotFoundError):
        _ = installer.version


def test_python_installer_exists(mocker: MockerFixture) -> None:
    mocker.patch(
        "poetry.utils.env.python.Python.find_poetry_managed_pythons",
        return_value=[
            mocker.Mock(
                implementation="cpython",
                free_threaded=False,
                version=Version.from_parts(3, 9, 1),
            )
        ],
    )
    installer = PythonInstaller(request="3.9.1")
    assert installer.exists()


def test_python_installer_does_not_exist(mocker: MockerFixture) -> None:
    mocker.patch(
        "poetry.utils.env.python.Python.find_poetry_managed_pythons", return_value=[]
    )
    installer = PythonInstaller(request="3.9.1")
    assert not installer.exists()


def test_python_installer_exists_with_bad_executables(mocker: MockerFixture) -> None:
    class BadPython:
        @property
        def implementation(self) -> str:
            return "cpython"

        @property
        def free_threaded(self) -> bool:
            return False

        @property
        def executable(self) -> Path:
            return cast("Path", mocker.Mock(as_posix=lambda: "/path/to/bad/python"))

        @property
        def version(self) -> None:
            raise CalledProcessError(1, "cmd")

    mocker.patch(
        "poetry.utils.env.python.Python.find_poetry_managed_pythons",
        return_value=[BadPython()],
    )

    installer = PythonInstaller(request="3.9.1")
    with pytest.raises(PoetryRuntimeError):
        assert not installer.exists()


@pytest.mark.parametrize(
    ("implementation", "requested", "expected"),
    [
        ("cpython", "pypy", False),
        ("pypy", "pypy", True),
    ],
)
def test_python_installer_exists_implementation(
    mocker: MockerFixture,
    implementation: Literal["cpython", "pypy"],
    requested: Literal["cpython", "pypy"],
    expected: bool,
) -> None:
    mocker.patch(
        "poetry.utils.env.python.Python.find_poetry_managed_pythons",
        return_value=[
            mocker.Mock(
                implementation=implementation,
                free_threaded=False,
                version=Version.from_parts(3, 9, 1),
            )
        ],
    )
    installer = PythonInstaller(request="3.9.1", implementation=requested)
    assert installer.exists() is expected


@pytest.mark.parametrize(
    ("free_threaded", "requested", "expected"),
    [
        ("cpython", "pypy", False),
        ("pypy", "pypy", True),
    ],
)
def test_python_installer_exists_free_threaded(
    mocker: MockerFixture, free_threaded: bool, requested: bool, expected: bool
) -> None:
    mocker.patch(
        "poetry.utils.env.python.Python.find_poetry_managed_pythons",
        return_value=[
            mocker.Mock(
                implementation="cpython",
                free_threaded=free_threaded,
                version=Version.from_parts(3, 9, 1),
            )
        ],
    )
    installer = PythonInstaller(request="3.9.1", free_threaded=requested)
    assert installer.exists() is expected


def test_python_installer_install(mocker: MockerFixture) -> None:
    mocker.patch(
        "pbs_installer.get_download_link",
        return_value=(mocker.Mock(major=3, minor=9, micro=1), None),
    )
    install = mocker.patch("pbs_installer.install")
    installer = PythonInstaller(request="3.9.1")
    installer.install()
    install.assert_called_once_with(
        "3.9.1",
        installer.installation_directory,
        version_dir=True,
        implementation="cpython",
        free_threaded=False,
    )


def test_python_installer_install_error(mocker: MockerFixture) -> None:
    mocker.patch("pbs_installer.install", side_effect=ValueError)
    installer = PythonInstaller(request="3.9.1")
    with pytest.raises(PythonInstallationError):
        installer.install()



from __future__ import annotations

import pytest

from packaging.utils import canonicalize_name
from poetry.core.packages.dependency_group import MAIN_GROUP
from poetry.core.version.markers import parse_marker

from poetry.packages.transitive_package_info import TransitivePackageInfo


DEV_GROUP = canonicalize_name("dev")


@pytest.mark.parametrize(
    "groups, expected",
    [
        (set(), "<empty>"),
        ({"main"}, 'sys_platform == "linux"'),
        ({"dev"}, 'python_version < "3.9"'),
        ({"main", "dev"}, 'sys_platform == "linux" or python_version < "3.9"'),
        ({"foo"}, "<empty>"),
        ({"main", "foo", "dev"}, 'sys_platform == "linux" or python_version < "3.9"'),
    ],
)
def test_get_marker(groups: set[str], expected: str) -> None:
    info = TransitivePackageInfo(
        depth=0,
        groups={MAIN_GROUP, DEV_GROUP},
        markers={
            MAIN_GROUP: parse_marker('sys_platform =="linux"'),
            DEV_GROUP: parse_marker('python_version < "3.9"'),
        },
    )
    assert str(info.get_marker([canonicalize_name(g) for g in groups])) == expected



from __future__ import annotations

import shutil

from typing import TYPE_CHECKING

import pytest

from cleo.io.null_io import NullIO

from poetry.factory import Factory
from poetry.publishing.uploader import Uploader
from poetry.publishing.uploader import UploadError


if TYPE_CHECKING:
    from pathlib import Path

    import responses

    from pytest_mock import MockerFixture

    from poetry.poetry import Poetry
    from tests.types import FixtureDirGetter


@pytest.fixture
def poetry(fixture_dir: FixtureDirGetter) -> Poetry:
    return Factory().create_poetry(fixture_dir("simple_project"))


@pytest.fixture
def uploader(poetry: Poetry) -> Uploader:
    return Uploader(poetry, NullIO())


@pytest.mark.parametrize(
    ("files", "expected_files", "expected_version"),
    [
        ([], [], ""),
        (
            ["simple_project-1.2.3.tar.gz", "simple_project-1.2.3-py3-none-any.whl"],
            ["simple_project-1.2.3.tar.gz", "simple_project-1.2.3-py3-none-any.whl"],
            "1.2.3",
        ),
        (  # other names are ignored
            [
                "simple_project-1.2.3.tar.gz",
                "other_project-1.2.3.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
                "other_project-1.2.3-py3-none-any.whl",
            ],
            ["simple_project-1.2.3.tar.gz", "simple_project-1.2.3-py3-none-any.whl"],
            "1.2.3",
        ),
        (  # older versions are ignored
            [
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.4.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
                "simple_project-1.2.4-py3-none-any.whl",
            ],
            ["simple_project-1.2.4.tar.gz", "simple_project-1.2.4-py3-none-any.whl"],
            "1.2.4",
        ),
        (  # older versions are ignored - only new sdist
            [
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.4.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
            ],
            ["simple_project-1.2.4.tar.gz"],
            "1.2.4",
        ),
        (  # older versions are ignored - only new wheel
            [
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
                "simple_project-1.2.4-py3-none-any.whl",
            ],
            ["simple_project-1.2.4-py3-none-any.whl"],
            "1.2.4",
        ),
        (  # older versions are ignored - local version
            [
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.3+hash1.tar.gz",
                "simple_project-1.2.3+hash2.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
                "simple_project-1.2.3+hash1-py3-none-any.whl",
                "simple_project-1.2.3+hash2-py3-none-any.whl",
            ],
            [
                "simple_project-1.2.3+hash2.tar.gz",
                "simple_project-1.2.3+hash2-py3-none-any.whl",
            ],
            "1.2.3+hash2",
        ),
        (  # older versions are ignore - pre-release
            [
                "simple_project-1.2.3rc1.tar.gz",
                "simple_project-1.2.3rc1-py3-none-any.whl",
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
            ],
            [
                "simple_project-1.2.3.tar.gz",
                "simple_project-1.2.3-py3-none-any.whl",
            ],
            "1.2.3",
        ),
    ],
)
def test_uploader_files_only_latest(
    poetry: Poetry,
    tmp_path: Path,
    files: list[str],
    expected_files: list[str],
    expected_version: str,
) -> None:
    for file in files:
        (tmp_path / file).touch()
    uploader = Uploader(poetry, NullIO(), dist_dir=tmp_path)

    assert uploader.files == [tmp_path / f for f in expected_files]
    assert uploader.version == expected_version


def test_uploader_properly_handles_400_errors(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    http.post("https://foo.com", status=400, body="Bad request")

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert str(e.value) == "HTTP Error 400: Bad Request | b'Bad request'"


def test_uploader_properly_handles_403_errors(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    http.post("https://foo.com", status=403, body="Unauthorized")

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert str(e.value) == "HTTP Error 403: Forbidden | b'Unauthorized'"


def test_uploader_properly_handles_nonstandard_errors(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    # content based off a true story.
    # Message changed to protect the ~~innocent~~ guilty.
    content = (
        b'{\n "errors": [ {\n '
        b'"status": 400,'
        b'"message": "I cant let you do that, dave"\n'
        b"} ]\n}"
    )
    http.post("https://foo.com", status=400, body=content)

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert str(e.value) == f"HTTP Error 400: Bad Request | {content!r}"


@pytest.mark.parametrize(
    ("status", "code"),
    [
        (308, "Permanent Redirect"),
        (307, "Temporary Redirect"),
        (304, "Not Modified"),
        (303, "See Other"),
        (302, "Found"),
        (301, "Moved Permanently"),
        (300, "Multiple Choices"),
    ],
)
def test_uploader_properly_handles_redirects(
    http: responses.RequestsMock, uploader: Uploader, status: int, code: str
) -> None:
    http.post("https://foo.com", status=status)

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert (
        str(e.value)
        == "Redirects are not supported. Is the URL missing a trailing slash?"
    )


def test_uploader_properly_handles_301_redirects(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    http.post("https://foo.com", status=301, body="Redirect")

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert (
        str(e.value)
        == "Redirects are not supported. Is the URL missing a trailing slash?"
    )


def test_uploader_registers_with_sdist_for_appropriate_400_errors(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    http.post("https://foo.com", status=400, body="No package was ever registered")

    with pytest.raises(UploadError):
        uploader.upload("https://foo.com")

    assert len(http.calls) == 2
    bodies = [c.request.body or b"" for c in http.calls]
    assert b'name=":action"\r\n\r\nfile_upload\r\n' in bodies[0]
    assert b'name=":action"\r\n\r\nsubmit\r\n' in bodies[1]
    assert b"sdist" in bodies[0]
    assert b"sdist" in bodies[1]
    assert b"bdist_wheel" not in bodies[0]
    assert b"bdist_wheel" not in bodies[1]


def test_uploader_register_uses_wheel_if_no_sdist(
    http: responses.RequestsMock, poetry: Poetry, tmp_path: Path
) -> None:
    dist_dir = tmp_path / "dist"
    dist_dir.mkdir()
    shutil.copy(
        poetry.file.path.parent / "dist" / "simple_project-1.2.3-py2.py3-none-any.whl",
        dist_dir,
    )

    uploader = Uploader(poetry, NullIO(), dist_dir=dist_dir)

    http.post("https://foo.com", status=400, body="No package was ever registered")

    with pytest.raises(UploadError):
        uploader.upload("https://foo.com")

    assert len(http.calls) == 2
    bodies = [c.request.body or b"" for c in http.calls]
    assert b'name=":action"\r\n\r\nfile_upload\r\n' in bodies[0]
    assert b'name=":action"\r\n\r\nsubmit\r\n' in bodies[1]
    assert b"sdist" not in bodies[0]
    assert b"sdist" not in bodies[1]
    assert b"bdist_wheel" in bodies[0]
    assert b"bdist_wheel" in bodies[1]


@pytest.mark.parametrize(
    "status, body",
    [
        (409, ""),
        (400, "File already exists"),
        (400, "Repository does not allow updating assets"),
        (400, "cannot be updated"),
        (403, "Not enough permissions to overwrite artifact"),
        (400, "file name has already been taken"),
    ],
)
def test_uploader_skips_existing(
    http: responses.RequestsMock, uploader: Uploader, status: int, body: str
) -> None:
    http.post("https://foo.com", status=status, body=body)

    # should not raise
    uploader.upload("https://foo.com", skip_existing=True)


def test_uploader_skip_existing_bubbles_unskippable_errors(
    http: responses.RequestsMock, uploader: Uploader
) -> None:
    http.post("https://foo.com", status=403, body="Unauthorized")

    with pytest.raises(UploadError):
        uploader.upload("https://foo.com", skip_existing=True)


def test_uploader_properly_handles_file_not_existing(
    mocker: MockerFixture, http: responses.RequestsMock, uploader: Uploader
) -> None:
    mocker.patch("pathlib.Path.is_file", return_value=False)

    with pytest.raises(UploadError) as e:
        uploader.upload("https://foo.com")

    assert f"Archive ({uploader.files[0]}) does not exist" == str(e.value)


def test_uploader_post_data_wheel(fixture_dir: FixtureDirGetter) -> None:
    file = (
        fixture_dir("simple_project")
        / "dist"
        / "simple_project-1.2.3-py2.py3-none-any.whl"
    )
    assert Uploader.post_data(file) == {
        "md5_digest": "fb4a5266406b9cf34ceaa88d1c8b7a01",
        "sha256_digest": "fc365a242d4de8b8661babc088f44b3df25e9e0017ef5dd7140dfe50f9323e16",
        "blake2_256_digest": "2e006d1fbfef0ed38fbded1ec1614dc4fd66f81061fe290528e2744dbc25ce31",
        "filetype": "bdist_wheel",
        "pyversion": "py2.py3",
        "metadata_version": "2.1",
        "name": "simple-project",
        "version": "1.2.3",
        "summary": "Some description.",
        "author": "Sébastien Eustace",
        "author_email": "sebastien@eustace.io",
        "license": "MIT",
        "classifiers": [
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
        "requires_python": ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
        "description": "My Package\n==========\n\n",
        "description_content_type": "text/x-rst",
        "keywords": "packaging, dependency, poetry",
        "home_page": "https://poetry.eustace.io",
        "project_urls": [
            "Documentation, https://poetry.eustace.io/docs",
            "Repository, https://github.com/sdispater/poetry",
        ],
    }


def test_uploader_post_data_sdist(fixture_dir: FixtureDirGetter) -> None:
    file = fixture_dir("simple_project") / "dist" / "simple_project-1.2.3.tar.gz"
    assert Uploader.post_data(file) == {
        "md5_digest": "e611cbb8f31258243d90f7681dfda68a",
        "sha256_digest": "c4a72becabca29ec2a64bf8c820bbe204d2268f53e102501ea5605bc1c1675d1",
        "blake2_256_digest": "d3df22f4944f6acd02105e7e2df61ef63c7b0f4337a12df549ebc2805a13c2be",
        "filetype": "sdist",
        "pyversion": "source",
        "metadata_version": "2.1",
        "name": "simple-project",
        "version": "1.2.3",
        "summary": "Some description.",
        "author": "Sébastien Eustace",
        "author_email": "sebastien@eustace.io",
        "classifiers": [
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
        "requires_python": ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
        "keywords": "packaging, dependency, poetry",
        "home_page": "https://poetry.eustace.io",
        "project_urls": [
            "Documentation, https://poetry.eustace.io/docs",
            "Repository, https://github.com/sdispater/poetry",
        ],
    }



from __future__ import annotations

import json
import re

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from urllib.parse import urlparse

import pytest
import responses

from packaging.utils import canonicalize_name

from poetry.repositories.legacy_repository import LegacyRepository
from tests.helpers import FIXTURE_PATH_REPOSITORIES_LEGACY
from tests.helpers import FIXTURE_PATH_REPOSITORIES_PYPI


if TYPE_CHECKING:
    from collections.abc import Callable

    from packaging.utils import NormalizedName
    from pytest import FixtureRequest
    from pytest_mock import MockerFixture
    from requests import PreparedRequest

    from poetry.repositories.link_sources.base import LinkSource
    from tests.types import HttpRequestCallback
    from tests.types import HttpResponse
    from tests.types import NormalizedNameTransformer
    from tests.types import SpecializedLegacyRepositoryMocker

pytest_plugins = [
    "tests.repositories.fixtures.python_hosted",
]


class TestLegacyRepository(LegacyRepository):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.json = False


@pytest.fixture
def legacy_repository_directory() -> Path:
    return FIXTURE_PATH_REPOSITORIES_LEGACY


@pytest.fixture
def legacy_package_json_locations() -> list[Path]:
    return [
        FIXTURE_PATH_REPOSITORIES_LEGACY / "json",
        FIXTURE_PATH_REPOSITORIES_PYPI / "json",
    ]


@pytest.fixture
def legacy_repository_package_names(legacy_repository_directory: Path) -> set[str]:
    return {
        package_html_file.stem
        for package_html_file in legacy_repository_directory.glob("*.html")
    }


@pytest.fixture
def legacy_repository_index_html(
    legacy_repository_directory: Path, legacy_repository_package_names: set[str]
) -> str:
    hrefs = [
        f'<a href="{name}/">{name}</a><br>' for name in legacy_repository_package_names
    ]

    return f"""<!DOCTYPE html>
    <html>
        <head>
            Legacy Repository
        </head>
        <body>
        {"".join(hrefs)}
        </body>
    </html>
    <!--TIMESTAMP 1709913893-->
    """


@pytest.fixture
def legacy_repository_index_json(
    legacy_repository_directory: Path, legacy_repository_package_names: set[str]
) -> dict[str, Any]:
    names = [{"name": name} for name in legacy_repository_package_names]
    return {"meta": {"api-version": "1.4"}, "projects": names}


@pytest.fixture
def legacy_repository_url() -> str:
    return "https://legacy.foo.bar"


@pytest.fixture
def legacy_repository_html_callback(
    legacy_repository_directory: Path,
    legacy_repository_index_html: str,
) -> HttpRequestCallback:
    def html_callback(request: PreparedRequest) -> HttpResponse:
        assert request.url
        if name := Path(urlparse(request.url).path).name:
            fixture = legacy_repository_directory / f"{name}.html"

            if not fixture.exists():
                return 404, {}, b"Not Found"

            return 200, {}, fixture.read_bytes()

        return 200, {}, legacy_repository_index_html.encode("utf-8")

    return html_callback


@pytest.fixture
def legacy_repository_json_callback(
    legacy_package_json_locations: list[Path],
    legacy_repository_index_json: dict[str, Any],
) -> HttpRequestCallback:
    def json_callback(request: PreparedRequest) -> HttpResponse:
        assert request.url
        headers = {"Content-Type": "application/vnd.pypi.simple.v1+json"}
        if name := Path(urlparse(request.url).path).name:
            fixture = Path()
            for location in legacy_package_json_locations:
                fixture = location / f"{name}.json"
                if fixture.exists():
                    break

            if not fixture.exists():
                return 404, {}, b"Not Found"

            return 200, headers, fixture.read_bytes()

        return 200, headers, json.dumps(legacy_repository_index_json).encode("utf-8")

    return json_callback


@pytest.fixture
def legacy_repository_html(
    http: responses.RequestsMock,
    legacy_repository_url: str,
    legacy_repository_html_callback: HttpRequestCallback,
    mock_files_python_hosted: None,
) -> TestLegacyRepository:
    http.add_callback(
        responses.GET,
        re.compile(r"^https://legacy\.(.*)+/?(.*)?$"),
        callback=legacy_repository_html_callback,
    )

    repo = TestLegacyRepository("legacy", legacy_repository_url, disable_cache=True)
    repo.json = False
    return repo


@pytest.fixture
def legacy_repository_json(
    http: responses.RequestsMock,
    legacy_repository_url: str,
    legacy_repository_json_callback: HttpRequestCallback,
    mock_files_python_hosted: None,
) -> TestLegacyRepository:
    http.add_callback(
        responses.GET,
        re.compile(r"^https://legacy\.(.*)+/?(.*)?$"),
        callback=legacy_repository_json_callback,
    )

    repo = TestLegacyRepository("legacy", legacy_repository_url, disable_cache=True)
    repo.json = True
    return repo


@pytest.fixture(params=["legacy_repository_html", "legacy_repository_json"])
def legacy_repository(request: FixtureRequest) -> TestLegacyRepository:
    return request.getfixturevalue(request.param)  # type: ignore[no-any-return]


@pytest.fixture
def specialized_legacy_repository_mocker(
    legacy_repository_html: LegacyRepository,
    legacy_repository_url: str,
    mocker: MockerFixture,
) -> SpecializedLegacyRepositoryMocker:
    """
    This is a mocker factory that allows tests cases to intercept and redirect to special case legacy html files by
    creating an instance of the mocked legacy repository and then mocking its get_page method for special cases.
    """

    def mock(
        transformer_or_suffix: NormalizedNameTransformer | str,
        repository_name: str = "special",
        repository_url: str = legacy_repository_url,
    ) -> LegacyRepository:
        specialized_repository = LegacyRepository(
            repository_name, repository_url, disable_cache=True
        )
        original_get_page = specialized_repository._get_page

        def _mocked_get_page(name: NormalizedName) -> LinkSource:
            return original_get_page(
                canonicalize_name(f"{name}{transformer_or_suffix}")
                if isinstance(transformer_or_suffix, str)
                else transformer_or_suffix(name)
            )

        mocker.patch.object(specialized_repository, "get_page", _mocked_get_page)

        return specialized_repository

    return mock


@pytest.fixture
def legacy_repository_with_extra_packages(
    specialized_legacy_repository_mocker: SpecializedLegacyRepositoryMocker,
) -> LegacyRepository:
    return specialized_legacy_repository_mocker("-with-extra-packages")


@pytest.fixture
def legacy_repository_partial_yank(
    specialized_legacy_repository_mocker: SpecializedLegacyRepositoryMocker,
) -> LegacyRepository:
    return specialized_legacy_repository_mocker("-partial-yank")


@pytest.fixture
def get_legacy_dist_url(legacy_repository_directory: Path) -> Callable[[str], str]:
    def get_url(name: str) -> str:
        package_name = name.split("-", 1)[0]
        path = legacy_repository_directory / f"{package_name}.html"
        if not path.exists():
            raise RuntimeError(
                f"Fixture for {package_name}.html not found in legacy fixtures"
            )
        content = path.read_text(encoding="utf-8")
        match = re.search(rf'<a href="(?P<url>[^"]+{re.escape(name)}[^"]*)"', content)
        if not match:
            raise RuntimeError(
                f"No URL for {name} found in legacy fixture {package_name}.html"
            )
        return match.group("url").split("#", 1)[0]

    return get_url


@pytest.fixture
def get_legacy_dist_size_and_upload_time(
    legacy_package_json_locations: list[Path],
) -> Callable[[str], tuple[int | None, str | None]]:
    def get_size_and_upload_time(name: str) -> tuple[int | None, str | None]:
        package_name = name.split("-", 1)[0]
        fixture_name = f"{package_name}.json"
        path = Path()
        for location in legacy_package_json_locations:
            path = location / fixture_name
            if path.exists():
                break
        if not path.exists():
            raise RuntimeError(
                f"Fixture for {fixture_name} not found in legacy fixtures"
            )
        with path.open("rb") as f:
            content = json.load(f)
        for file in content["files"]:
            if file["filename"] == name:
                size = file.get("size")
                upload_time = file.get("upload-time")
                return int(size) if size else None, upload_time
        raise RuntimeError(f"No URL for {name} found in legacy fixture {fixture_name}")

    return get_size_and_upload_time
